"""config I/O functions."""

import abc
import codecs
from ast import literal_eval
from configparser import ConfigParser
from os.path import abspath, dirname, exists, join, splitext
from pathlib import Path
from typing import Dict, List, Optional, Union

import yaml
from tomli import load as load_toml
from tomli_w import dump as dump_toml

__all__ = [
    "configread",
    "configwrite",
]


def _process_config_out(d):
    ret = {}
    if isinstance(d, dict):
        for k, v in d.items():
            if v is None:
                ret[k] = "NONE"
            else:
                ret[k] = _process_config_out(v)
    else:
        ret = d

    return ret


def _process_config_in(d):
    ret = {}
    if isinstance(d, dict):
        for k, v in d.items():
            if v == "NONE":
                ret[k] = None
            else:
                ret[k] = _process_config_in(v)
    else:
        ret = d

    return ret


def configread(
    config_fn: Union[Path, str],
    defaults: Optional[Dict] = None,
    abs_path: bool = False,
    skip_abspath_sections: Optional[List] = None,
    **kwargs,
) -> Dict:
    """Read configuration/workflow file and parse to (nested) dictionary.

    Parameters
    ----------
    config_fn : Union[Path, str]
        Path to configuration file
    defaults : dict, optional
        Nested dictionary with default options, by default dict()
    abs_path : bool, optional
        If True, parse string values to an absolute path if the a file or folder
        with that name (string value) relative to the config file exist,
        by default False
    skip_abspath_sections: list, optional
        These sections are not evaluated for absolute paths if abs_path=True,
        by default ['update_config']
    **kwargs
        Additional keyword arguments that are passed to the read_ini`
        function.

    Returns
    -------
    cfdict : dict
        Configuration dictionary.
    """
    defaults = defaults or {}
    skip_abspath_sections = skip_abspath_sections or ["setup_config"]
    # read
    ext = splitext(config_fn)[-1].strip()
    if ext in [".yaml", ".yml"]:
        with open(config_fn, "rb") as f:
            cfdict = yaml.safe_load(f)
        cfdict = _process_config_in(cfdict)
    elif ext == ".toml":  # user defined
        with open(config_fn, "rb") as f:
            cfdict = load_toml(f)
        cfdict = _process_config_in(cfdict)
    else:
        cfdict = read_ini_config(config_fn, **kwargs)
    # parse absolute paths
    if abs_path:
        root = Path(dirname(config_fn))
        cfdict = parse_abspath(cfdict, root, skip_abspath_sections)

    # update defaults
    if defaults:
        _cfdict = defaults.copy()
        _cfdict.update(cfdict)
        cfdict = _cfdict
    return cfdict


def configwrite(config_fn: Union[str, Path], cfdict: dict, **kwargs) -> None:
    """Write configuration/workflow dictionary to file.

    Parameters
    ----------
    config_fn : Union[Path, str]
        Path to configuration file
    cfdict : dict
        Configuration dictionary. If the configuration contains headers,
        the first level keys are the section headers, the second level
        option-value pairs.
    encoding : str, optional
        File encoding, by default "utf-8"
    cf : ConfigParser, optional
        Alternative configuration parser, by default None
    noheader : bool, optional
        Set true for a single-level configuration dictionary with no headers,
        by default False
    **kwargs
        Additional keyword arguments that are passed to the `write_ini_config`
        function.
    """
    root = Path(dirname(config_fn))
    _cfdict = parse_relpath(cfdict.copy(), root)
    ext = splitext(config_fn)[-1].strip()
    if ext in [".yaml", ".yml"]:
        _cfdict = _process_config_out(_cfdict)  # should not be done for ini
        with open(config_fn, "w") as f:
            yaml.dump(_cfdict, f, sort_keys=False)
    elif ext == ".toml":  # user defined
        _cfdict = _process_config_out(_cfdict)
        with open(config_fn, "wb") as f:
            dump_toml(_cfdict, f)
    else:
        write_ini_config(config_fn, _cfdict, **kwargs)


def read_ini_config(
    config_fn: Union[Path, str],
    encoding: str = "utf-8",
    cf: ConfigParser = None,
    skip_eval: bool = False,
    skip_eval_sections: Optional[list] = None,
    noheader: bool = False,
) -> dict:
    """Read configuration ini file and parse to (nested) dictionary.

    Parameters
    ----------
    config_fn : Union[Path, str]
        Path to configuration file
    encoding : str, optional
        File encoding, by default "utf-8"
    cf : ConfigParser, optional
        Alternative configuration parser, by default None
    skip_eval : bool, optional
        If True, do not evaluate string values, by default False
    skip_eval_sections : list, optional
        These sections are not evaluated for string values
        if skip_eval=True, by default []
    noheader : bool, optional
        Set true for a single-level configuration file with no headers, by default False

    Returns
    -------
    cfdict : dict
        Configuration dictionary.
    """
    skip_eval_sections = skip_eval_sections or []
    if cf is None:
        cf = ConfigParser(allow_no_value=True, inline_comment_prefixes=[";", "#"])
    elif isinstance(cf, abc.ABCMeta):  # not yet instantiated
        cf = cf()
    cf.optionxform = str  # preserve capital letter
    with codecs.open(config_fn, "r", encoding=encoding) as fp:
        cf.read_file(fp)
        cfdict = cf._sections
    # parse values
    cfdict = parse_values(cfdict, skip_eval, skip_eval_sections)
    # add dummy header
    if noheader and "dummy" in cfdict:
        cfdict = cfdict["dummy"]
    return cfdict


def write_ini_config(
    config_fn: Union[Path, str],
    cfdict: dict,
    encoding: str = "utf-8",
    cf: ConfigParser = None,
    noheader: bool = False,
) -> None:
    """Write configuration dictionary to ini file.

    Parameters
    ----------
    config_fn : Union[Path, str]
        Path to configuration file
    cfdict : dict
        Configuration dictionary.
    encoding : str, optional
        File encoding, by default "utf-8"
    cf : ConfigParser, optional
        Alternative configuration parser, by default None
    noheader : bool, optional
        Set true for a single-level configuration dictionary with no headers,
        by default False
    """
    if cf is None:
        cf = ConfigParser(allow_no_value=True, inline_comment_prefixes=[";", "#"])
    elif isinstance(cf, abc.ABCMeta):  # not yet instantiated
        cf = cf()
    cf.optionxform = str  # preserve capital letter
    if noheader:  # add dummy header
        cfdict = {"dummy": cfdict}
    cf.read_dict(cfdict)
    with codecs.open(config_fn, "w", encoding=encoding) as fp:
        cf.write(fp)


def parse_relpath(cfdict: dict, root: Path) -> dict:
    """Parse string/path value to relative path if possible."""

    def _relpath(value, root):
        if isinstance(value, str) and str(Path(value)).startswith(str(root)):
            value = Path(value)
        if isinstance(value, Path):
            try:
                rel_path = value.relative_to(root)
                value = str(rel_path).replace("\\", "/")
            except ValueError:
                pass  # `value` path is not relative to root
        return value

    # loop through n-level of dict
    for key, val in cfdict.items():
        if isinstance(val, dict):
            cfdict[key] = parse_relpath(val, root)
        else:
            cfdict[key] = _relpath(val, root)
    return cfdict


def parse_abspath(
    cfdict: dict, root: Path, skip_abspath_sections: Optional[List] = None
) -> dict:
    """Parse string value to absolute path from config file."""
    skip_abspath_sections = skip_abspath_sections or ["setup_config"]

    def _abspath(value, root):
        if exists(join(root, value)):
            value = Path(abspath(join(root, value)))
        return value

    # loop through n-level of dict
    for key, val in cfdict.items():
        if isinstance(val, dict):
            if key not in skip_abspath_sections:
                cfdict[key] = parse_abspath(val, root)
        elif isinstance(val, list) and all([isinstance(v, str) for v in val]):
            cfdict[key] = [_abspath(v, root) for v in val]
        elif isinstance(val, str):
            cfdict[key] = _abspath(val, root)
    return cfdict


def parse_values(
    cfdict: dict,
    skip_eval: bool = False,
    skip_eval_sections: Optional[List] = None,
):
    """Parse string values to python default objects.

    Parameters
    ----------
    cfdict : dict
        Configuration dictionary.
    skip_eval : bool, optional
        Set true to skip evaluation, by default False
    skip_eval_sections : List, optional
        List of sections to skip evaluation, by default []

    Returns
    -------
    cfdict : dict
        Configuration dictionary with evaluated values.
    """
    skip_eval_sections = skip_eval_sections or []
    # loop through two-level dict: section, key-value pairs
    for section in cfdict:
        # evaluate ini items to parse to python default objects:
        if skip_eval or section in skip_eval_sections:
            cfdict[section].update(
                {key: str(var) for key, var in cfdict[section].items()}
            )  # cast None type values to str
            continue  # do not evaluate
        # numbers, tuples, lists, dicts, sets, booleans, and None
        for key, value in cfdict[section].items():
            try:
                value = literal_eval(value)
            except Exception:
                pass
            if isinstance(value, str) and len(value) == 0:
                value = None
            cfdict[section].update({key: value})
    return cfdict



"""Implementations for river workflows."""
import logging
from typing import Union

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from pyflwdir import Flwdir, FlwdirRaster
from scipy import ndimage

from ..gis_utils import spread2d

logger = logging.Logger(__name__)

__all__ = ["river_width", "river_depth"]


def river_width(
    gdf_stream: gpd.GeoDataFrame,
    da_rivmask: xr.DataArray,
    nmin=5,
) -> np.ndarray:
    """Return average river width along a segment based on a river mask raster.

    For each segment in gdf_stream the associated area is calculated from stream mask
    and divided by the segment length to obtain the average width.

    Parameters
    ----------
    gdf_stream : gpd.GeoDataFrame
        River segments
    da_rivmask : xr.DataArray
        Boolean river mask in projected grid.
    nmin : int, optional
        Minimum number of cells in rivmask to calculate the width, by default 5

    Returns
    -------
    rivwth : np.ndarray
        Average width per segment in gdf_stream
    """
    assert da_rivmask.raster.crs.is_projected
    gdf_stream = gdf_stream.copy()
    # get/check river length
    if "rivlen" not in gdf_stream.columns:
        gdf_stream["rivlen"] = gdf_stream.to_crs(da_rivmask.raster.crs).length
    # rasterize streams
    gdf_stream["segid"] = np.arange(1, gdf_stream.index.size + 1, dtype=np.int32)
    segid = da_rivmask.raster.rasterize(gdf_stream, "segid").astype(np.int32)
    segid.raster.set_nodata(0)
    segid.name = "segid"
    # remove islands to get total width of braided rivers
    da_mask = da_rivmask.copy()
    da_mask.data = ndimage.binary_fill_holes(da_mask.values)
    # find nearest stream segment for all river cells
    segid_spread = spread2d(segid, da_mask)
    # get average width based on da_rivmask area and segment length
    cellarea = abs(np.multiply(*da_rivmask.raster.res))
    seg_count = ndimage.sum(
        da_rivmask, segid_spread["segid"].values, gdf_stream["segid"].values
    )
    rivwth = seg_count * cellarea / gdf_stream["rivlen"]
    valid = np.logical_and(gdf_stream["rivlen"] > 0, seg_count > nmin)
    return np.where(valid, rivwth, -9999)


def river_depth(
    data: Union[xr.Dataset, pd.DataFrame, gpd.GeoDataFrame],
    method: str,
    flwdir: Union[Flwdir, FlwdirRaster] = None,
    min_rivdph: float = 1.0,
    manning: float = 0.03,
    qbankfull_name: str = "qbankfull",
    rivwth_name: str = "rivwth",
    rivzs_name: str = "rivzs",
    rivdst_name: str = "rivdst",
    rivslp_name: str = "rivslp",
    rivman_name: str = "rivman",
    **kwargs,
) -> Union[xr.DataArray, np.ndarray]:
    """Derive river depth estimates based bankfull discharge.

    Parameters
    ----------
    data : xr.Dataset, pd.DataFrame, gpd.GeoDataFrame
        Dataset/DataFrame containing required variables
    method : {'powlaw', 'manning', 'gvf'}
        Method to estimate the river depth:

        * powlaw [1]_ [2]_: power-law hc*Qbf**hp, requires bankfull discharge (Qbf).
          Optionally, `hc` (default = 0.27) and `hp` (default = 0.30) set in `kwargs`.
        * manning [3]_: river depth for kinematic conditions,
          requires bankfull discharge, river width, river slope in `data`;
          the river manning roughness either in data or as constant and
          optionally `min_rivslp` (default = 1e-5) set in `kwargs`.
        * gvf [4]_: gradually varying flow, requires bankfull discharge,
          river width, river surface elevation in `data`;
          the river manning roughness either in data
          or as constant and optionally `min_rivslp` (default = 1e-5) set in `kwargs`.
    flwdir : Flwdir, FlwdirRaster, optional
        Flow directions, required if method is not powlaw
    min_rivdph : float, optional
        Minimum river depth [m], by default 1.0
    manning : float, optional
        Constant manning roughness [s/m^{1/3}] used if `rivman_name` not in data,
        by default 0.03
    qbankfull_name, rivwth_name, rivzs_name, rivdst_name, rivslp_name, rivman_name:
        Name for variables in data: bankfull discharge [m3/s], river width [m],
        bankfull water surface elevation profile [m+REF], distance to river outlet [m],
        river slope [m/m] and river manning roughness [s/m^{1/3}]
    kwargs:
        additional arguments to be passed down

    Returns
    -------
    rivdph: xr.DataArray, np.ndarray
        River depth [m]. A DataArray is returned if the input data is a Dataset,
        otherwise a array with the shape of one input data variable is returned.

    References
    ----------
    .. [1] Leopold & Maddock (1953). The hydraulic geometry of stream channels and some
            physiographic implications (No. 252; Professional Paper).
            U.S. Government Printing Office. https://doi.org/10.3133/pp252
    .. [2] Andreadis et al. (2013). A simple global river bankfull width and depth
            database. Water Resources Research, 49(10), 7164-7168. https://doi.org/10.1002/wrcr.20440
    .. [3] Sampson et al. (2015). A high-resolution global flood hazard model.
            Water Resources Research, 51(9), 7358-7381. https://doi.org/10.1002/2015WR016954
    .. [4] Neal et al. (2021). Estimating river channel bathymetry in large scale
            flood inundation models. Water Resources Research, 57(5). https://doi.org/10.1029/2020wr028301

    See Also
    --------
    pyflwdir.FlwdirRaster.river_depth
    """
    methods = ["powlaw", "manning", "gvf"]
    if method == "powlaw":

        def rivdph_powlaw(qbankfull, hc=0.27, hp=0.30, min_rivdph=1.0):
            return np.maximum(hc * qbankfull**hp, min_rivdph)

        rivdph = rivdph_powlaw(data[qbankfull_name], min_rivdph=min_rivdph, **kwargs)
    elif method in ["manning", "gvf"]:
        assert flwdir is not None
        rivdph = flwdir.river_depth(
            qbankfull=data[qbankfull_name].values,
            rivwth=data[rivwth_name].values,
            zs=data[rivzs_name].values if rivzs_name in data else None,
            rivdst=data[rivdst_name].values if rivdst_name in data else None,
            rivslp=data[rivslp_name].values if rivslp_name in data else None,
            manning=data[rivman_name].values if rivman_name in data else manning,
            method=method,
            min_rivdph=min_rivdph,
            **kwargs,
        )
    else:
        raise ValueError(f"Method unknown {method}, select from {methods}")
    if isinstance(data, xr.Dataset):
        rivdph = xr.DataArray(
            dims=data.raster.dims, coords=data.raster.coords, data=rivdph
        )
        rivdph.raster.set_nodata(-9999.0)
        rivdph.raster.set_crs(data.raster.crs)
    return rivdph



# -*- coding: utf-8 -*-
"""HydroMT NetworkModel class definition."""

import logging
from typing import List

import xarray as xr

from .model_api import Model

__all__ = ["NetworkModel"]
logger = logging.getLogger(__name__)


class NetworkModel(Model):

    """Implementation for the network models."""

    _CLI_ARGS = {"region": "setup_region"}
    _NAME = "network_model"

    def __init__(
        self,
        root: str = None,
        mode: str = "w",
        config_fn: str = None,
        data_libs: List[str] = None,
        logger=logger,
    ):
        """Initialize a NetworkModel for network models."""
        super().__init__(
            root=root,
            mode=mode,
            config_fn=config_fn,
            data_libs=data_libs,
            logger=logger,
        )

        # placeholders
        # TODO decide on data type
        self._network = xr.Dataset()  # xr.Dataset representation of all network data

    def read(
        self,
        components: List = None,
    ) -> None:
        """Read the complete model schematization and configuration from model files.

        Parameters
        ----------
        components : List, optional
            List of model components to read, each should have an associated
            read_<component> method. By default ['config', 'maps',
            'network', 'geoms', 'tables', 'forcing', 'states', 'results']
        """
        components = components or [
            "config",
            "network",
            "geoms",
            "tables",
            "forcing",
            "states",
            "results",
        ]
        super().read(components=components)

    def write(
        self,
        components: List = None,
    ) -> None:
        """Write the complete model schematization and configuration to model files.

        Parameters
        ----------
        components : List, optional
            List of model components to write, each should have an
            associated write_<component> method. By default ['config', 'maps',
            'network', 'geoms', 'tables', 'forcing', 'states']
        """
        components = components or [
            "config",
            "network",
            "geoms",
            "tables",
            "forcing",
            "states",
        ]
        super().write(components=components)

    # TODO: make NetworkMixin class with following properties/methods
    @property
    def network(self):  # noqa: D102
        raise NotImplementedError()

    def set_network(self):  # noqa: D102
        raise NotImplementedError()

    def read_network(self):  # noqa: D102
        raise NotImplementedError()

    def write_network(self):  # noqa: D102
        raise NotImplementedError()



"""Functions to handle nodata values."""
from enum import Enum
from logging import Logger

from .exceptions import NoDataException


class NoDataStrategy(Enum):
    """Strategy to handle nodata values."""

    RAISE = "raise"
    IGNORE = "ignore"


def _exec_nodata_strat(msg: str, strategy: NoDataStrategy, logger: Logger) -> None:
    """Execute nodata strategy."""
    if strategy == NoDataStrategy.RAISE:
        raise NoDataException(msg)
    elif strategy == NoDataStrategy.IGNORE:
        logger.warning(msg)
    else:
        raise NotImplementedError(f"NoDataStrategy '{strategy}' not implemented.")



"""Implementation for the Pandas Dataframe adapter."""
import warnings
from datetime import datetime
from logging import Logger, getLogger
from os.path import join
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
import pandas as pd
from pystac import Asset as StacAsset
from pystac import Catalog as StacCatalog
from pystac import Item as StacItem

from hydromt.nodata import NoDataException, NoDataStrategy, _exec_nodata_strat
from hydromt.typing import ErrorHandleMethod, StrPath, TimeRange, Variables

from .data_adapter import DataAdapter

logger = getLogger(__name__)

__all__ = [
    "DataFrameAdapter",
]


class DataFrameAdapter(DataAdapter):

    """DataAdapter implementation for Pandas Dataframes."""

    _DEFAULT_DRIVER = "csv"
    _DRIVERS = {"xlsx": "excel", "xls": "excel"}

    def __init__(
        self,
        path: StrPath,
        driver: Optional[str] = None,
        filesystem: Optional[str] = None,
        nodata: Optional[Union[dict, float, int]] = None,
        rename: Optional[dict] = None,
        unit_mult: Optional[dict] = None,
        unit_add: Optional[dict] = None,
        meta: Optional[dict] = None,
        attrs: Optional[dict] = None,
        driver_kwargs: Optional[dict] = None,
        storage_options: Optional[dict] = None,
        name: str = "",  # optional for now
        catalog_name: str = "",  # optional for now
        provider: Optional[str] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        """Initiate data adapter for 2D tabular data.

        This object contains all properties required to read supported files into
        a :py:func:`pandas.DataFrame`.
        In addition it keeps meta data to be able to reproduce which data is used.

        Parameters
        ----------
        path: str, Path
            Path to data source. If the dataset consists of multiple files, the path may
            contain {variable}, {year}, {month} placeholders as well as path
            search pattern using a '*' wildcard.
        driver: {'csv', 'parquet', 'xlsx', 'xls', 'fwf'}, optional
            Driver to read files with, for 'csv' :py:func:`~pandas.read_csv`,
            for 'parquet' :py:func:`~pandas.read_parquet`, for {'xlsx', 'xls'}
            :py:func:`~pandas.read_excel`, and for 'fwf' :py:func:`~pandas.read_fwf`.
            By default the driver is inferred from the file extension and falls back to
            'csv' if unknown.
        filesystem: str, optional
            Filesystem where the data is stored (local, cloud, http etc.).
            If None (default) the filesystem is inferred from the path.
            See :py:func:`fsspec.registry.known_implementations` for all options.
        nodata: dict, float, int, optional
            Missing value number. Only used if the data has no native missing value.
            Nodata values can be differentiated between variables using a dictionary.
        rename: dict, optional
            Mapping of native data source variable to output source variable name as
            required by hydroMT.
        unit_mult, unit_add: dict, optional
            Scaling multiplication and addition to change to map from the native
            data unit to the output data unit as required by hydroMT.
        meta: dict, optional
            Metadata information of dataset, prefably containing the following keys:
            {'source_version', 'source_url', 'source_license',
            'paper_ref', 'paper_doi', 'category'}
        placeholders: dict, optional
            Placeholders to expand yaml entry to multiple entries (name and path)
            based on placeholder values
        attrs: dict, optional
            Additional attributes relating to data variables. For instance unit
            or long name of the variable.
        extent: None
            Not used in this adapter. Only here for compatability with other adapters.
        driver_kwargs, dict, optional
            Additional key-word arguments passed to the driver.
        storage_options: dict, optional
            Additional key-word arguments passed to the fsspec FileSystem object.
        name, catalog_name: str, optional
            Name of the dataset and catalog, optional for now.
        """
        driver_kwargs = driver_kwargs or {}
        attrs = attrs or {}
        meta = meta or {}
        unit_add = unit_add or {}
        unit_mult = unit_mult or {}
        rename = rename or {}
        storage_options = storage_options or {}

        if kwargs:
            warnings.warn(
                "Passing additional keyword arguments to be used by the "
                "DataFrameAdapter driver is deprecated and will be removed "
                "in a future version. Please use 'driver_kwargs' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            driver_kwargs.update(kwargs)
        super().__init__(
            path=path,
            driver=driver,
            filesystem=filesystem,
            nodata=nodata,
            rename=rename,
            unit_mult=unit_mult,
            unit_add=unit_add,
            meta=meta,
            attrs=attrs,
            driver_kwargs=driver_kwargs,
            storage_options=storage_options,
            name=name,
            catalog_name=catalog_name,
            provider=provider,
            version=version,
        )

    def to_file(
        self,
        data_root: StrPath,
        data_name: str,
        driver: Optional[str] = None,
        variables: Optional[Variables] = None,
        time_tuple: Optional[TimeRange] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        logger: Logger = logger,
        **kwargs,
    ) -> Optional[Tuple[StrPath, str, Dict[str, Any]]]:
        """Save a dataframe slice to a file.

        Parameters
        ----------
        data_root : str or Path
            Path to the output folder.
        data_name : str
            Name of the output file without extension.
        driver : str, optional
            Driver to write the file, e.g., 'csv','parquet', 'excel'. If None,
            the default behavior is used.
        variables : list of str, optional
            Names of DataFrame columns to include in the output. By default,
            all columns are included.
        time_tuple : tuple of str or datetime, optional
            Start and end date of the period of interest. By default, the entire time
            period of the DataFrame is included.
        **kwargs : dict
            Additional keyword arguments to be passed to the file writing method.

        Returns
        -------
        fn_out : str
            Absolute path to the output file.
        driver : str
            Name of the driver used to read the data.
            See :py:func:`~hydromt.data_catalog.DataCatalog.get_geodataset`.
        kwargs: dict
            The additional keyword arguments that were passed in.


        """
        kwargs.pop("bbox", None)
        obj = self.get_data(
            time_tuple=time_tuple,
            variables=variables,
            handle_nodata=handle_nodata,
            logger=logger,
        )

        if obj is None:
            return None

        read_kwargs = dict()
        if driver is None or driver == "csv":
            # always write as CSV
            driver = "csv"
            fn_out = join(data_root, f"{data_name}.csv")
            obj.to_csv(fn_out, **kwargs)
            read_kwargs["index_col"] = 0
        elif driver == "parquet":
            fn_out = join(data_root, f"{data_name}.parquet")
            obj.to_parquet(fn_out, **kwargs)
        elif driver == "excel":
            fn_out = join(data_root, f"{data_name}.xlsx")
            obj.to_excel(fn_out, **kwargs)
        else:
            raise ValueError(f"DataFrame: Driver {driver} is unknown.")

        return fn_out, driver, read_kwargs

    def get_data(
        self,
        variables: Optional[Variables] = None,
        time_tuple: Optional[TimeRange] = None,
        logger: Logger = logger,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
    ) -> Optional[pd.DataFrame]:
        """Return a DataFrame.

        Returned data is optionally sliced by time and variables,
        based on the properties of this DataFrameAdapter. For a detailed
        description see: :py:func:`~hydromt.data_catalog.DataCatalog.get_dataframe`
        """
        try:
            # load data
            fns = self._resolve_paths(variables=variables)
            self.mark_as_used()  # mark used
            df = self._read_data(fns, logger=logger)
            # just raise an exxceptoin so we can handle the strategy in one place (the except)
            if df is None:
                raise NoDataException()
            # rename variables and parse nodata
            df = self._rename_vars(df)
            df = self._set_nodata(df)
            # slice data
            df = DataFrameAdapter._slice_data(
                df,
                variables,
                time_tuple,
                logger=logger,
            )
            if df is None:
                raise NoDataException()

            # uniformize data
            df = self._apply_unit_conversion(df, logger=logger)
            df = self._set_metadata(df)
            return df
        except NoDataException:
            _exec_nodata_strat(
                f"No data was read from source: {self.name}",
                strategy=handle_nodata,
                logger=logger,
            )
            return None

    def _read_data(
        self,
        fns: List[StrPath],
        logger: Logger = logger,
    ) -> Optional[pd.DataFrame]:
        if len(fns) > 1:
            raise ValueError(
                f"DataFrame: Reading multiple {self.driver} files is not supported."
            )
        kwargs = self.driver_kwargs.copy()
        path = fns[0]
        logger.info(f"Reading {self.name} {self.driver} data from {self.path}")
        if self.driver in ["csv"]:
            df = pd.read_csv(path, **kwargs)
        elif self.driver == "parquet":
            _ = kwargs.pop("index_col", None)
            df = pd.read_parquet(path, **kwargs)
        elif self.driver in ["xls", "xlsx", "excel"]:
            df = pd.read_excel(path, engine="openpyxl", **kwargs)
        elif self.driver in ["fwf"]:
            df = pd.read_fwf(path, **kwargs)
        else:
            raise IOError(f"DataFrame: driver {self.driver} unknown.")

        # cast is just for the type checkers, it doesn't actually do anything
        df = cast(pd.DataFrame, df)

        if len(df) == 0:
            return None
        else:
            return df

    def _rename_vars(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.rename:
            rename = {k: v for k, v in self.rename.items() if k in df.columns}
            df = df.rename(columns=rename)
        return df

    def _set_nodata(self, df: pd.DataFrame) -> pd.DataFrame:
        # parse nodata values
        cols = df.select_dtypes([np.number]).columns
        if self.nodata is not None and len(cols) > 0:
            if not isinstance(self.nodata, dict):
                nodata = {c: self.nodata for c in cols}
            else:
                nodata = self.nodata
            for c in cols:
                mv = nodata.get(c, None)
                if mv is not None:
                    is_nodata = np.isin(df[c], np.atleast_1d(mv))
                    df[c] = np.where(is_nodata, np.nan, df[c])
        return df

    @staticmethod
    def _slice_data(
        df: pd.DataFrame,
        variables: Optional[Variables] = None,
        time_tuple: Optional[TimeRange] = None,
        logger: Logger = logger,
    ) -> Optional[pd.DataFrame]:
        """Return a sliced DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            the dataframe to be sliced.
        variables : list of str, optional
            Names of DataFrame columns to include in the output. By default all columns
        time_tuple : tuple of str, datetime, optional
            Start and end date of period of interest. By default the entire time period
            of the dataset is returned.

        Returns
        -------
        pd.DataFrame
            Tabular data
        """
        if variables is not None:
            variables = np.atleast_1d(variables).tolist()
            if np.any([var not in df.columns for var in variables]):
                raise ValueError(f"DataFrame: Not all variables found: {variables}")
            df = df.loc[:, variables]

        if time_tuple is not None and np.dtype(df.index).type == np.datetime64:
            logger.debug(f"Slicing time dime {time_tuple}")
            try:
                df = df[df.index.slice_indexer(*time_tuple)]
            except IndexError:
                df = pd.DataFrame()

        if len(df) == 0:
            return None
        else:
            return df

    def _apply_unit_conversion(
        self, df: pd.DataFrame, logger: Logger = logger
    ) -> pd.DataFrame:
        unit_names = list(self.unit_mult.keys()) + list(self.unit_add.keys())
        unit_names = [k for k in unit_names if k in df.columns]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} columns.")
        for name in list(set(unit_names)):  # unique
            m = self.unit_mult.get(name, 1)
            a = self.unit_add.get(name, 0)
            df[name] = df[name] * m + a
        return df

    def _set_metadata(self, df: pd.DataFrame) -> pd.DataFrame:
        df.attrs.update(self.meta)

        # set column attributes
        for col in self.attrs:
            if col in df.columns:
                df[col].attrs.update(**self.attrs[col])

        return df

    def to_stac_catalog(
        self,
        on_error: ErrorHandleMethod = ErrorHandleMethod.COERCE,
    ) -> Optional[StacCatalog]:
        """
        Convert a dataframe into a STAC Catalog representation.

        The collection will contain an asset for each of the associated files.


        Parameters
        ----------
        - on_error (str, optional): The error handling strategy.
          Options are: "raise" to raise an error on failure, "skip" to skip the
          dataset on failure, and "coerce" (default) to set default values on failure.

        Returns
        -------
        - Optional[StacCatalog]: The STAC Catalog representation of the dataset, or None
          if the dataset was skipped.
        """
        if on_error == ErrorHandleMethod.SKIP:
            logger.warning(
                f"Skipping {self.name} during stac conversion because"
                "because detecting temporal extent failed."
            )
            return
        elif on_error == ErrorHandleMethod.COERCE:
            stac_catalog = StacCatalog(
                self.name,
                description=self.name,
            )
            stac_item = StacItem(
                self.name,
                geometry=None,
                bbox=[0, 0, 0, 0],
                properties=self.meta,
                datetime=datetime(1, 1, 1),
            )
            stac_asset = StacAsset(str(self.path))
            stac_item.add_asset("hydromt_path", stac_asset)

            stac_catalog.add_item(stac_item)
            return stac_catalog
        else:
            raise NotImplementedError(
                "DataframeAdapter does not support full stac conversion as it lacks"
                " spatio-temporal dimensions"
            )



"""HydroMT exceptions."""


class NoDataException(Exception):
    """Exception raised for errors in the input.

    Attributes
    ----------
        message -- explanation of the error
    """

    def __init__(self, message="No data available"):
        self.message = message
        super().__init__(self.message)



import sys

from packaging.version import Version

__all__ = []

HAS_XUGRID = False
HAS_SHAPELY20 = False
HAS_PYET = False
HAS_GCSFS = False
HAS_S3FS = False
HAS_OPENPYXL = False
HAS_RIO_VRT = False

try:
    from shapely import __version__ as SH_VERSION

    if Version(SH_VERSION) >= Version("2.0.0"):
        HAS_SHAPELY20 = True
except ImportError:
    pass

try:
    import openpyxl

    HAS_OPENPYXL = True
except ImportError:
    pass

try:
    import xugrid

    HAS_XUGRID = True
except ImportError:
    pass

try:
    import gcsfs

    HAS_GCSFS = True

except ImportError:
    pass

try:
    import s3fs

    HAS_S3FS = True

except ImportError:
    pass

try:
    import pyet

    HAS_PYET = True

except ModuleNotFoundError:
    pass


try:
    import rio_vrt

    HAS_RIO_VRT = True

except ModuleNotFoundError:
    pass


# entrypoints in standard library only compatible from 3.10 onwards
py_version = sys.version_info
if py_version[0] >= 3 and py_version[1] >= 10:
    from importlib.metadata import entry_points, Distribution, EntryPoint, EntryPoints  # noqa: I001
else:
    from importlib_metadata import entry_points, Distribution, EntryPoint, EntryPoints  # noqa: I001



"""Implementations related to logging."""

import logging
import logging.handlers
import os
import sys
from functools import wraps

from . import __version__

FMT = "%(asctime)s - %(name)s - %(module)s - %(levelname)s - %(message)s"


def setuplog(
    name: str = "hydromt",
    path: str = None,
    log_level: int = 20,
    fmt: str = FMT,
    append: bool = True,
) -> logging.Logger:
    """Set up the logging on sys.stdout and file if path is given.

    Parameters
    ----------
    name : str, optional
        logger name, by default "hydromt"
    path : str, optional
        path to logfile, by default None
    log_level : int, optional
        Log level [0-50], by default 20 (info)
    fmt : str, optional
        log message formatter
    append : bool, optional
        Whether to append (True) or overwrite (False) to a logfile at path,
        by default True

    Returns
    -------
    logging.Logger
        _description_
    """
    logger = logging.getLogger(name)
    for _ in range(len(logger.handlers)):
        logger.handlers.pop().close()  # remove and close existing handlers
    logging.captureWarnings(True)
    logger.setLevel(log_level)
    console = logging.StreamHandler(sys.stdout)
    console.setLevel(log_level)
    console.setFormatter(logging.Formatter(fmt))
    logger.addHandler(console)
    if path is not None:
        if append is False and os.path.isfile(path):
            os.unlink(path)
        add_filehandler(logger, path, log_level=log_level, fmt=fmt)
    logger.info(f"HydroMT version: {__version__}")

    return logger


def add_filehandler(logger, path, log_level=20, fmt=FMT):
    """Add file handler to logger."""
    if not os.path.isdir(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    isfile = os.path.isfile(path)
    ch = logging.FileHandler(path)
    ch.setFormatter(logging.Formatter(fmt))
    ch.setLevel(log_level)
    logger.addHandler(ch)
    if isfile:
        logger.debug(f"Appending log messages to file {path}.")
    else:
        logger.debug(f"Writing log messages to new file {path}.")


def logged(logger):
    """Define a decorator that logs the execution of a function.

    This decorator logs the input arguments, output, and any raised exceptions of
    the decorated function using the provided logger.

    Parameters
    ----------
    logger : Logger
        The logger object used to log the function execution.

    Returns
    -------
    wrap : callable
        The decorator function that wraps the decorated function.
    """

    def wrap(function):
        @wraps(function)
        def wrapper(*args, **kwargs):
            f = function.__name__
            logger.debug(f"{f} - args={args} kwargs={kwargs}")
            try:
                response = function(*args, **kwargs)
            except Exception as error:
                e = error.__class__.__name__
                emsg = str(error)
                logger.error(f"{f} - raised {e} with error '{emsg}'")
                raise
            logger.debug(f"{f} - return={response}")
            return response

        return wrapper

    return wrap



"""Implementation of necessary statistical methods."""

import warnings

import bottleneck
import numpy as np
import xarray as xr

warnings.filterwarnings("ignore")


# PERFORMANCE METRICS
def bias(sim, obs, dim="time"):
    r"""Return the bias between two time series.

        .. math::
         Bias=\\frac{1}{N}\\sum_{i=1}^{N}(obs_{i}-sim_{i})

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        bias
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    bias = xr.apply_ufunc(_bias, sim, obs, **kwargs)
    bias.name = "bias"
    return bias


def percentual_bias(sim, obs, dim="time"):
    r"""Return the percentual bias between two time series.

        .. math::
         PBias= 100 * \\frac{\\sum_{i=1}^{N}(sim_{i}-obs_{i})}{\\sum_{i=1}^{N}(obs_{i})}

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        percentual bias
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    pbias = xr.apply_ufunc(_pbias, sim, obs, **kwargs)
    pbias.name = "pbias"
    return pbias


def volumetric_error(sim, obs, dim="time"):
    r"""Return the volumetric error between two time series.

        .. math::
         VE= 1 - \\frac{\\sum_{i=1}^{N}(|sim_{i}-obs_{i}|)}{\\sum_{i=1}^{N}(obs_{i})}

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        volumetric error
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    ve = xr.apply_ufunc(_ve, sim, obs, **kwargs)
    ve.name = "ve"
    return ve


def nashsutcliffe(sim, obs, dim="time"):
    r"""Return the Nash-Sutcliffe model efficiency.

    Efficiency is based on a simulated and observed time series.

        .. math::
         NSE = 1-\\frac{\\sum_{i=1}^{N}(sim_{i}-obs_{i})^2}
         {\\sum_{i=1}^{N}(obs_{i}-\\bar{obs})^2}

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        the Nash-Sutcliffe model efficiency
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    nse = xr.apply_ufunc(_nse, sim, obs, **kwargs)
    nse.name = "nse"
    return nse


def lognashsutcliffe(sim, obs, epsilon=1e-6, dim="time"):
    r"""Return the log Nash-Sutcliffe model efficiency.

    Efficiency calculation is based on simulated and observed time series.

        .. math::
         NSE = 1-\\frac{\\sum_{i=1}^{N}(log(sim_{i})-log(obs_{i}))^2}
         {\\sum_{i=1}^{N}(log(sim_{i})-log(\\bar{obs})^2}-1)*-1

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    epsilon : float, optional
        small value to avoid taking the log of zero (the default is 1e-6)
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        the log of the Nash-Sutcliffe model efficiency
    """
    obs = np.log(obs + epsilon)
    sim = np.log(sim + epsilon)
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    log_nse = xr.apply_ufunc(_nse, sim, obs, **kwargs)
    log_nse.name = "log_nse"
    return log_nse


def pearson_correlation(sim, obs, dim="time"):
    """Return the Pearson correlation coefficient of two time series.

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        the pearson correlation coefficient
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    pearsonr = xr.apply_ufunc(_pearson_correlation, sim, obs, **kwargs)
    pearsonr.name = "pearson_coef"
    return pearsonr


def spearman_rank_correlation(sim, obs, dim="time"):
    """Return the spearman rank correlation coefficient of two time series.

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        the spearman rank correlation
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    spearmanr = xr.apply_ufunc(_spearman_correlation, sim, obs, **kwargs)
    spearmanr.name = "spearmanr_coef"
    return spearmanr


def kge_non_parametric(sim, obs, dim="time"):
    """Return the Non Parametric Kling-Gupta Efficiency (KGE, 2018).

    Calculation is based on the two time series with decomposed scores.

    .. ref:

        Pool, Vis, and Seibert, 2018 Evaluating model performance: towards
        a non-parametric variant of the Kling-Gupta efficiency,
        Hydrological Sciences Journal.

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataSet
        Non Parametric Kling-Gupta Efficiency (2018) and with decomposed score
    """
    cc = spearman_rank_correlation(sim, obs, dim=dim)
    cc.name = "kge_np_spearman_rank_correlation_coef"
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    alpha = xr.apply_ufunc(_fdc_alpha, sim, obs, **kwargs)
    alpha.name = "kge_np_rel_var"
    beta = sim.sum(dim=dim) / obs.sum(dim=dim)
    beta.name = "kge_np_bias"
    kge = 1 - np.sqrt((cc - 1) ** 2 + (alpha - 1) ** 2 + (beta - 1) ** 2)
    kge.name = "kge_np"
    ds_out = xr.merge([kge, cc, alpha, beta])
    return ds_out


def kge_non_parametric_flood(sim, obs, dim="time"):
    """Return the Non Parametric Kling-Gupta Efficiency (KGE, 2018) of two time series.

    this KGE is optimized for flood peaks using Pearson (see Pool et al., 2018).

    .. ref:

        Pool, Vis, and Seibert, 2018 Evaluating model performance: towards
        a non-parametric variant of the Kling-Gupta efficiency,
        Hydrological Sciences Journal.

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataSet
        Non Parametric Kling-Gupta Efficiency (2018) optimize for flood peaks
        using pearson (see Pool et al., 2018) and with decomposed score
    """
    cc = pearson_correlation(sim, obs, dim=dim)
    cc.name = "kge_np_flood_pearson_coef"
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    alpha = xr.apply_ufunc(_fdc_alpha, sim, obs, **kwargs)
    alpha.name = "kge_np_flood_rel_var"
    beta = sim.sum(dim=dim) / obs.sum(dim=dim)
    beta.name = "kge_np_flood_bias"
    kge = 1 - np.sqrt((cc - 1) ** 2 + (alpha - 1) ** 2 + (beta - 1) ** 2)
    kge.name = "kge_np_flood"
    ds_out = xr.merge([kge, cc, alpha, beta])
    return ds_out


def rsquared(sim, obs, dim="time"):
    r"""Return the coefficient of determination of two time series.

        .. math::
         r^2=(\\frac{\\sum ^n _{i=1}(e_i - \\bar{e})(s_i - \\bar{s})}
         {\\sqrt{\\sum ^n _{i=1}(e_i - \\bar{e})^2}
         \\sqrt{\\sum ^n _{i=1}(s_i - \\bar{s})^2}})^2

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        the coefficient of determination
    """
    # R2 equals the square of the Pearson correlation coefficient between obs and sim
    rsquared = pearson_correlation(sim, obs, dim=dim) ** 2
    rsquared.name = "rsquared"
    return rsquared


def mse(sim, obs, dim="time"):
    r"""Return the mean squared error (MSE) between two time series.

        .. math::
         MSE=\\frac{1}{N}\\sum_{i=1}^{N}(e_{i}-s_{i})^2

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        the mean squared error
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    mse = xr.apply_ufunc(_mse, sim, obs, **kwargs)
    mse.name = "mse"
    return mse


def rmse(sim, obs, dim="time"):
    r"""Return the root mean squared error between two time series.

        .. math::
         RMSE=\\sqrt{\\frac{1}{N}\\sum_{i=1}^{N}(e_{i}-s_{i})^2}

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataArray
        the root mean squared error
    """
    rmse = np.sqrt(mse(sim, obs, dim=dim))
    rmse.name = "rmse"
    return rmse


def rsr(sim, obs, dim="time"):
    r"""Return the RMSE-observations standard deviation (RSR) between two time series.

        .. math::
         RSR=\\frac{RMSE}{STDEVobs}

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series

    Returns
    -------
    xarray DataArray
        the root squared error
    """
    # wrap numpy function
    kwargs = dict(
        input_core_dims=[[dim], [dim]], dask="parallelized", output_dtypes=[float]
    )
    rsr = xr.apply_ufunc(_rsr, sim, obs, **kwargs)
    rsr.name = "rsr"
    return rsr


def kge(sim, obs, dim="time"):
    r"""Return the Kling-Gupta Efficiency (KGE) of two time series.

    .. ref:

        Gupta, Kling, Yilmaz, Martinez, 2009, Decomposition of the mean
        squared error and NSE performance criteria: Implications for improving
        hydrological modelling.

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataSet
        Kling-Gupta Efficiency and with decomposed scores
    """
    cc = pearson_correlation(sim, obs, dim=dim)
    cc.name = "kge_pearson_coef"
    alpha = sim.std(dim=dim) / obs.std(dim=dim)
    alpha.name = "kge_rel_var"
    beta = sim.sum(dim=dim) / obs.sum(dim=dim)
    beta.name = "kge_bias"
    kge = 1 - np.sqrt((cc - 1) ** 2 + (alpha - 1) ** 2 + (beta - 1) ** 2)
    kge.name = "kge"
    ds_out = xr.merge([kge, cc, alpha, beta])
    return ds_out


def kge_2012(sim, obs, dim="time"):
    r"""Return the Kling-Gupta Efficiency (KGE, 2012) of two time series.

    .. ref:

        Kling, H., Fuchs, M., & Paulin, M. (2012). Runoff conditions in the
        upper Danube basin under an ensemble of climate change scenarios.
        Journal of Hydrology, 424, 264-277, doi:10.1016/j.jhydrol.2012.01.011.

    Parameters
    ----------
    sim : xarray DataArray
        simulations time series
    obs : xarray DataArray
        observations time series
    dim : str, optional
        name of time dimension in sim and obs (the default is 'time')

    Returns
    -------
    xarray DataSet
        Kling-Gupta Efficiency (2012) and with decomposed scores
    """
    cc = pearson_correlation(sim, obs, dim=dim)
    cc.name = "kge_2012_pearson_coef"
    beta = sim.sum(dim=dim) / obs.sum(dim=dim)
    beta.name = "kge_2012_bias"
    # divide alpha by bias
    alpha = (sim.std(dim=dim) * obs.sum(dim=dim)) / (
        obs.std(dim=dim) * sim.sum(dim=dim)
    )
    alpha.name = "kge_2012_rel_var"
    kge_2012 = 1 - np.sqrt((cc - 1) ** 2 + (alpha - 1) ** 2 + (beta - 1) ** 2)
    kge_2012.name = "kge_2012"
    ds_out = xr.merge([kge_2012, cc, alpha, beta])
    return ds_out


# correlation ufunc function
# from http://xarray.pydata.org/en/stable/dask.html#automatic-parallelization
def _covariance(x, y):
    return np.nanmean(
        (x - np.nanmean(x, axis=-1, keepdims=True))
        * (y - np.nanmean(y, axis=-1, keepdims=True)),
        axis=-1,
    )


def _pearson_correlation(x, y):
    return _covariance(x, y) / (np.nanstd(x, axis=-1) * np.nanstd(y, axis=-1))


def _spearman_correlation(x, y):
    x_ranks = bottleneck.nanrankdata(x, axis=-1)
    y_ranks = bottleneck.nanrankdata(y, axis=-1)
    return _pearson_correlation(x_ranks, y_ranks)


# numpy functions
def _fdc_alpha(sim, obs, axis=-1):
    fdc_s = np.sort(sim, axis=axis) / (np.nanmean(sim, axis=axis) * len(sim))
    fdc_o = np.sort(obs, axis=axis) / (np.nanmean(obs, axis=axis) * len(obs))
    return 1 - 0.5 * np.nansum(np.abs(fdc_s - fdc_o), axis=axis)


def _bias(sim, obs, axis=-1):
    """Bias."""
    return np.nansum(sim - obs, axis=axis) / np.nansum(np.isfinite(obs), axis=axis)


def _pbias(sim, obs, axis=-1):
    """Percentual bias."""
    return 100 * np.nansum(sim - obs, axis=axis) / np.nansum(obs, axis=axis)


def _ve(sim, obs, axis=-1):
    """Volumetric Error."""
    ve = 1 - np.nansum(np.absolute(sim - obs), axis=axis) / np.nansum(obs, axis=axis)
    return ve


def _mse(sim, obs, axis=-1):
    """Mean squared error."""
    mse = np.nansum((obs - sim) ** 2, axis=axis) / np.nansum(
        np.isfinite(obs), axis=axis
    )
    return mse


def _rsr(sim, obs, axis=-1):
    """RMSE-observations standard deviation."""
    rmse = np.sqrt(_mse(sim, obs, axis=axis))
    stdev = np.sqrt(
        np.nansum((obs - np.nanmean(obs, axis=axis)) ** 2, axis=axis)
        / np.nansum(np.isfinite(obs), axis=axis)
    )
    return rmse / stdev


def _nse(sim, obs, axis=-1):
    """nash-sutcliffe efficiency."""
    obs_mean = np.nanmean(obs, axis=axis)
    a = np.nansum((sim - obs) ** 2, axis=axis)
    b = np.nansum((obs - obs_mean[..., None]) ** 2, axis=axis)
    return 1 - a / b



"""Implementation for the geodataset DataAdapter."""
import warnings
from datetime import datetime
from logging import Logger, getLogger
from os.path import basename, splitext
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import pyproj
import xarray as xr
from pyproj.exceptions import CRSError
from pystac import Asset as StacAsset
from pystac import Catalog as StacCatalog
from pystac import Item as StacItem
from pystac import MediaType

from hydromt.nodata import NoDataException, NoDataStrategy, _exec_nodata_strat
from hydromt.typing import (
    Bbox,
    Data,
    ErrorHandleMethod,
    GeoDatasetSource,
    Geom,
    GeomBuffer,
    Predicate,
    StrPath,
    TimeRange,
    TotalBounds,
    Variables,
)
from hydromt.utils import has_no_data

from .. import gis_utils, io
from ..raster import GEO_MAP_COORD
from .data_adapter import DataAdapter
from .utils import netcdf_writer, shift_dataset_time, zarr_writer

logger = getLogger(__name__)

__all__ = ["GeoDatasetAdapter", "GeoDatasetSource"]


class GeoDatasetAdapter(DataAdapter):

    """DatasetAdapter for GeoDatasets."""

    _DEFAULT_DRIVER = "vector"
    _DRIVERS = {
        "nc": "netcdf",
    }

    def __init__(
        self,
        path: StrPath,
        driver: Optional[str] = None,
        filesystem: Optional[str] = None,
        crs: Optional[Union[int, str, dict]] = None,
        nodata: Optional[Union[dict, float, int]] = None,
        rename: Optional[dict] = None,
        unit_mult: Optional[dict] = None,
        unit_add: Optional[dict] = None,
        meta: Optional[dict] = None,
        attrs: Optional[dict] = None,
        extent: Optional[dict] = None,
        driver_kwargs: Optional[dict] = None,
        storage_options: Optional[dict] = None,
        name: str = "",
        catalog_name: str = "",
        provider: Optional[str] = None,
        version: Optional[str] = None,
        **kwargs,
    ):
        """Initiate data adapter for geospatial timeseries data.

        This object contains all properties required to read supported files into
        a single unified GeoDataset, i.e. :py:class:`xarray.Dataset` with geospatial
        point geometries. In addition it keeps meta data to be able to reproduce which
        data is used.

        Parameters
        ----------
        path: str, Path
            Path to data source. If the dataset consists of multiple files, the path may
            contain {variable}, {year}, {month} placeholders as well as path
            search pattern using a ``*`` wildcard.
        driver: {'vector', 'netcdf', 'zarr'}, optional
            Driver to read files with,
            for 'vector' :py:func:`~hydromt.io.open_geodataset`,
            for 'netcdf' :py:func:`xarray.open_mfdataset`.
            By default the driver is inferred from the file extension and falls back to
            'vector' if unknown.
        filesystem: str, optional
            Filesystem where the data is stored (local, cloud, http etc.).
            If None (default) the filesystem is inferred from the path.
            See :py:func:`fsspec.registry.known_implementations` for all options.
        crs: int, dict, or str, optional
            Coordinate Reference System. Accepts EPSG codes (int or str);
            proj (str or dict) or wkt (str). Only used if the data has no native CRS.
        nodata: float, int, optional
            Missing value number. Only used if the data has no native missing value.
            Nodata values can be differentiated between variables using a dictionary.
        rename: dict, optional
            Mapping of native data source variable to output source variable name as
            required by hydroMT.
        unit_mult, unit_add: dict, optional
            Scaling multiplication and addition to change to map from the native
            data unit to the output data unit as required by hydroMT.
        meta: dict, optional
            Metadata information of dataset, prefably containing the following keys:
            - 'source_version'
            - 'source_url'
            - 'source_license'
            - 'paper_ref'
            - 'paper_doi'
            - 'category'
        placeholders: dict, optional
            Placeholders to expand yaml entry to multiple entries (name and path)
            based on placeholder values
        attrs: dict, optional
            Additional attributes relating to data variables. For instance unit
            or long name of the variable.
        extent: Extent(typed dict), Optional
            Dictionary describing the spatial and time range the dataset covers.
            should be of the form:
            -  "bbox": [xmin, ymin, xmax, ymax],
            -  "time_range": [start_datetime, end_datetime],
            data, and time_range should be inclusive on both sides.
        driver_kwargs, dict, optional
            Additional key-word arguments passed to the driver.
        storage_options: dict, optional
            Additional key-word arguments passed to the fsspec FileSystem object.
        name, catalog_name: str, optional
            Name of the dataset and catalog, optional.
        provider: str, optional
            A name to identifiy the specific provider of the dataset requested.
            if None is provided, the last added source will be used.
        version: str, optional
            A name to identifiy the specific version of the dataset requested.
            if None is provided, the last added source will be used.
        """
        driver_kwargs = driver_kwargs or {}
        extent = extent or {}
        if kwargs:
            warnings.warn(
                "Passing additional keyword arguments to be used by the "
                "GeoDatasetAdapter driver is deprecated and will be removed "
                "in a future version. Please use 'driver_kwargs' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            driver_kwargs.update(kwargs)
        super().__init__(
            path=path,
            driver=driver,
            filesystem=filesystem,
            nodata=nodata,
            rename=rename,
            unit_mult=unit_mult,
            unit_add=unit_add,
            meta=meta,
            attrs=attrs,
            driver_kwargs=driver_kwargs,
            storage_options=storage_options,
            name=name,
            catalog_name=catalog_name,
            provider=provider,
            version=version,
        )
        self.crs = crs
        self.extent = extent

    def to_file(
        self,
        data_root: StrPath,
        data_name: str,
        bbox: Optional[Bbox] = None,
        time_tuple: Optional[TimeRange] = None,
        variables: Optional[List[str]] = None,
        driver: Optional[str] = None,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        logger: Logger = logger,
        **kwargs,
    ) -> Optional[Tuple[StrPath, Optional[str], Dict[str, Any]]]:
        """Save a data slice to file.

        Parameters
        ----------
        data_root : str, Path
            Path to output folder
        data_name : str
            Name of output file without extension.
        bbox : array-like of floats
            (xmin, ymin, xmax, ymax) bounding box of area of interest.
        time_tuple : tuple of str, datetime, optional
            Start and end date of period of interest. By default the entire time period
            of the dataset is returned.
        driver : str, optional
            Driver to write file, e.g.: 'netcdf', 'zarr', by default None
        variables : list of str, optional
            Names of GeoDataset variables to return. By default all dataset variables
            are returned.
        **kwargs
            Additional keyword arguments that are passed to the `to_zarr`
            function.

        Returns
        -------
        fn_out: str
            Absolute path to output file
        driver: str
            Name of driver to read data with, see
            :py:func:`~hydromt.data_catalog.DataCatalog.get_geodataset`
        """
        obj = self.get_data(
            bbox=bbox,
            time_tuple=time_tuple,
            variables=variables,
            handle_nodata=handle_nodata,
            logger=logger,
            single_var_as_array=variables is None,
        )

        # we'll handle the correct strategy in the except clause
        if obj is None:
            return None

        read_kwargs = {}

        # much better for mem/storage/processing if dtypes are set correctly
        for name, coord in obj.coords.items():
            if coord.values.dtype != object:
                continue

            # not sure if coordinates values of different dtypes
            # are possible, but let's just hope users aren't
            # that mean for now.
            if isinstance(coord.values[0], str):
                obj[name] = obj[name].astype(str)

        if driver is None or driver == "netcdf":
            # always write netcdf
            fn_out = netcdf_writer(
                obj=obj,
                data_root=data_root,
                data_name=data_name,
                variables=variables,
            )
        elif driver == "zarr":
            fn_out = zarr_writer(
                obj=obj, data_root=data_root, data_name=data_name, **kwargs
            )
        else:
            raise ValueError(f"GeoDataset: Driver {driver} unknown.")

        return fn_out, driver, read_kwargs

    def get_data(
        self,
        bbox: Optional[Bbox] = None,
        geom: Optional[Geom] = None,
        buffer: Optional[GeomBuffer] = 0,
        predicate: Predicate = "intersects",
        variables: Optional[Variables] = None,
        time_tuple: Optional[TimeRange] = None,
        single_var_as_array: bool = True,
        handle_nodata: NoDataStrategy = NoDataStrategy.RAISE,
        logger: Logger = logger,
    ) -> Optional[Data]:
        """Return a clipped, sliced and unified GeoDataset.

        For a detailed description see:
        :py:func:`~hydromt.data_catalog.DataCatalog.get_geodataset`
        """
        try:
            # load data
            fns = self._resolve_paths(variables=variables, time_tuple=time_tuple)
            self.mark_as_used()  # mark used
            ds = self._read_data(fns, logger=logger)

            if ds is None:
                raise NoDataException()

            # rename variables and parse data and attrs
            ds = self._rename_vars(ds)
            ds = self._validate_spatial_coords(ds)
            ds = self._set_crs(ds, logger=logger)
            ds = self._set_nodata(ds)
            ds = self._shift_time(ds, logger=logger)
            # slice
            ds = GeoDatasetAdapter._slice_data(
                ds,
                variables,
                geom,
                bbox,
                buffer,
                predicate,
                time_tuple,
                logger=logger,
            )

            if ds is None:
                raise NoDataException()

            # uniformize
            ds = self._apply_unit_conversion(ds, logger=logger)
            ds = self._set_metadata(ds)
            # return array if single var and single_var_as_array
            return self._single_var_as_array(ds, single_var_as_array, variables)
        except NoDataException:
            _exec_nodata_strat(
                f"No data was read from source: {self.name}",
                strategy=handle_nodata,
                logger=logger,
            )
            return None

    def _read_data(
        self,
        fns: List[StrPath],
        logger: Logger = logger,
    ) -> Data:
        kwargs = self.driver_kwargs.copy()
        if len(fns) > 1 and self.driver in ["vector", "zarr"]:
            raise ValueError(
                f"GeoDataset: Reading multiple {self.driver} files is not supported."
            )
        logger.info(f"Reading {self.name} {self.driver} data from {self.path}")
        if self.driver in ["netcdf"]:
            ds = xr.open_mfdataset(fns, **kwargs)
        elif self.driver == "zarr":
            ds = xr.open_zarr(fns[0], **kwargs)
        elif self.driver == "vector":
            ds = io.open_geodataset(fn_locs=fns[0], crs=self.crs, **kwargs)
        else:
            raise ValueError(f"GeoDataset: Driver {self.driver} unknown")

        if has_no_data(ds):
            raise NoDataException()

        return ds

    def _rename_vars(self, ds: Data) -> Data:
        rm = {k: v for k, v in self.rename.items() if k in ds}
        ds = ds.rename(rm)
        return ds

    def _validate_spatial_coords(self, ds: Data) -> Data:
        if GEO_MAP_COORD in ds.data_vars:
            ds = ds.set_coords(GEO_MAP_COORD)
        try:
            ds.vector.set_spatial_dims()
            idim = ds.vector.index_dim
            if idim not in ds:  # set coordinates for index dimension if missing
                ds[idim] = xr.IndexVariable(idim, np.arange(ds.dims[idim]))
            coords = [ds.vector.x_name, ds.vector.y_name, idim]
            coords = [item for item in coords if item is not None]
            ds = ds.set_coords(coords)
        except ValueError:
            raise ValueError(
                f"GeoDataset: No spatial geometry dimension found in data {self.path}"
            )
        return ds

    def _set_crs(self, ds: Data, logger: Logger = logger) -> Data:
        # set crs
        if ds.vector.crs is None and self.crs is not None:
            ds.vector.set_crs(self.crs)
        elif ds.vector.crs is None:
            raise ValueError(
                f"GeoDataset {self.name}: CRS not defined in data catalog or data."
            )
        elif self.crs is not None and ds.vector.crs != pyproj.CRS.from_user_input(
            self.crs
        ):
            logger.warning(
                f"GeoDataset {self.name}: CRS from data catalog does not match CRS of"
                " data. The original CRS will be used. Please check your data catalog."
            )
        return ds

    @staticmethod
    def _slice_data(
        ds: Data,
        variables: Optional[Variables] = None,
        geom: Optional[Geom] = None,
        bbox: Optional[Bbox] = None,
        buffer: Optional[GeomBuffer] = 0,
        predicate: Predicate = "intersects",
        time_tuple: Optional[TimeRange] = None,
        logger: Logger = logger,
    ) -> Optional[Data]:
        """Slice the dataset in space and time.

        Arguments
        ---------
        ds : xarray.Dataset or xarray.DataArray
            The GeoDataset to slice.
        variables : str or list of str, optional.
            Names of variables to return.
        geom : geopandas.GeoDataFrame/Series,
            A geometry defining the area of interest.
        bbox : array-like of floats
            (xmin, ymin, xmax, ymax) bounding box of area of interest
            (in WGS84 coordinates).
        buffer : float, optional
            Buffer distance [m] applied to the geometry or bbox. By default 0 m.
        predicate : str, optional
            Predicate used to filter the GeoDataFrame, see
            :py:func:`hydromt.gis_utils.filter_gdf` for details.
        time_tuple : tuple of str, datetime, optional
            Start and end date of period of interest. By default the entire time period
            of the dataset is returned.

        Returns
        -------
        ds : xarray.Dataset
            The sliced GeoDataset.
        """
        if isinstance(ds, xr.DataArray):
            if ds.name is None:
                # dummy name, required to create dataset
                # renamed to variable in _single_var_as_array
                ds.name = "data"
            ds = ds.to_dataset()
        elif variables is not None:
            variables = np.atleast_1d(variables).tolist()
            if len(variables) > 1 or len(ds.data_vars) > 1:
                mvars = [var not in ds.data_vars for var in variables]
                if any(mvars):
                    raise ValueError(f"GeoDataset: variables not found {mvars}")
                ds = ds[variables]
        if time_tuple is not None:
            ds = GeoDatasetAdapter._slice_temporal_dimension(
                ds, time_tuple, logger=logger
            )
        if geom is not None or bbox is not None:
            ds = GeoDatasetAdapter._slice_spatial_dimension(
                ds,
                geom,
                bbox,
                buffer,
                predicate,
                logger=logger,
            )
        if has_no_data(ds):
            return None
        else:
            return ds

    @staticmethod
    def _slice_spatial_dimension(
        ds: Data,
        geom: Optional[Geom],
        bbox: Optional[Bbox],
        buffer: Optional[GeomBuffer],
        predicate: Predicate,
        logger: Logger = logger,
    ) -> Optional[Data]:
        geom = gis_utils.parse_geom_bbox_buffer(geom, bbox, buffer)
        bbox_str = ", ".join([f"{c:.3f}" for c in geom.total_bounds])
        epsg = geom.crs.to_epsg()
        logger.debug(f"Clip {predicate} [{bbox_str}] (EPSG:{epsg})")
        ds = ds.vector.clip_geom(geom, predicate=predicate)
        if has_no_data(ds):
            return None
        else:
            return ds

    def _shift_time(self, ds: Data, logger: Logger = logger) -> Data:
        dt = self.unit_add.get("time", 0)
        return shift_dataset_time(dt=dt, ds=ds, logger=logger)

    @staticmethod
    def _slice_temporal_dimension(
        ds: Data,
        time_tuple: Optional[TimeRange],
        logger: Logger = logger,
    ) -> Optional[Data]:
        if (
            "time" in ds.dims
            and ds["time"].size > 1
            and np.issubdtype(ds["time"].dtype, np.datetime64)
        ):
            logger.debug(f"Slicing time dim {time_tuple}")
            ds = ds.sel(time=slice(*time_tuple))
        if has_no_data(ds):
            return None
        else:
            return ds

    def _set_metadata(self, ds: Data) -> Data:
        if self.attrs:
            if isinstance(ds, xr.DataArray):
                ds.attrs.update(self.attrs[ds.name])
            else:
                for k in self.attrs:
                    ds[k].attrs.update(self.attrs[k])

        ds.attrs.update(self.meta)
        return ds

    def _set_nodata(self, ds: Data) -> Data:
        if self.nodata is not None:
            if not isinstance(self.nodata, dict):
                nodata = {k: self.nodata for k in ds.data_vars.keys()}
            else:
                nodata = self.nodata
            for k in ds.data_vars:
                mv = nodata.get(k, None)
                if mv is not None and ds[k].vector.nodata is None:
                    ds[k].vector.set_nodata(mv)
        return ds

    def _apply_unit_conversion(self, ds: Data, logger: Logger = logger) -> Data:
        unit_names = list(self.unit_mult.keys()) + list(self.unit_add.keys())
        unit_names = [k for k in unit_names if k in ds.data_vars]
        if len(unit_names) > 0:
            logger.debug(f"Convert units for {len(unit_names)} variables.")
        for name in list(set(unit_names)):  # unique
            m = self.unit_mult.get(name, 1)
            a = self.unit_add.get(name, 0)
            da = ds[name]
            attrs = da.attrs.copy()
            nodata_isnan = da.vector.nodata is None or np.isnan(da.vector.nodata)
            # nodata value is explicitly set to NaN in case no nodata value is provided
            nodata = np.nan if nodata_isnan else da.vector.nodata
            data_bool = ~np.isnan(da) if nodata_isnan else da != nodata
            ds[name] = xr.where(data_bool, da * m + a, nodata)
            ds[name].attrs.update(attrs)  # set original attributes
        return ds

    def get_bbox(
        self,
        detect: bool = True,
    ) -> Optional[TotalBounds]:
        """Return the bounding box and espg code of the dataset.

        if the bounding box is not set and detect is True,
        :py:meth:`hydromt.GeoDatasetAdapter.detect_bbox` will be used to detect it.

        Parameters
        ----------
        detect: bool, Optional
            whether to detect the bounding box if it is not set. If False, and it's not
            set None will be returned.

        Returns
        -------
        bbox: Tuple[np.float64,np.float64,np.float64,np.float64]
            the bounding box coordinates of the data. coordinates are returned as
            [xmin,ymin,xmax,ymax]
        crs: int
            The ESPG code of the CRS of the coordinates returned in bbox
        """
        bbox = self.extent.get("bbox", None)
        if bbox is None and detect:
            bbox, crs = self.detect_bbox()

        crs = self.crs

        return bbox, crs

    def get_time_range(
        self,
        detect: bool = True,
    ) -> Optional[TimeRange]:
        """Detect the time range of the dataset.

        if the time range is not set and detect is True,
        :py:meth:`hydromt.GeoDatasetAdapter.detect_time_range` will be used
        to detect it.


        Parameters
        ----------
        detect: bool, Optional
            whether to detect the time range if it is not set. If False, and it's not
            set None will be returned.

        Returns
        -------
        range: Tuple[np.datetime64, np.datetime64]
            A tuple containing the start and end of the time dimension. Range is
            inclusive on both sides.
        """
        time_range = self.extent.get("time_range", None)
        if time_range is None and detect:
            time_range = self.detect_time_range()

        return time_range

    def detect_bbox(
        self,
        ds: Optional[Data] = None,
    ) -> Optional[TotalBounds]:
        """Detect the bounding box and crs of the dataset.

        If no dataset is provided, it will be fetched according to the settings in the
        adapter. also see :py:meth:`hydromt.GeoDatasetAdapter.get_data`. the
        coordinates are in the CRS of the dataset itself, which is also returned
        alongside the coordinates.


        Parameters
        ----------
        ds: xr.Dataset, xr.DataArray, Optional
            the dataset to detect the bounding box of.
            If none is provided, :py:meth:`hydromt.GeoDatasetAdapter.get_data`
            will be used to fetch the it before detecting.

        Returns
        -------
        bbox: Tuple[np.float64,np.float64,np.float64,np.float64]
            the bounding box coordinates of the data. coordinates are returned as
            [xmin,ymin,xmax,ymax]
        crs: int
            The ESPG code of the CRS of the coordinates returned in bbox
        """
        if ds is None:
            ds = self.get_data()

        crs = ds.vector.crs.to_epsg()
        bounds = ds.vector.bounds
        return bounds, crs

    def detect_time_range(
        self,
        ds: Optional[Data] = None,
    ) -> Optional[TimeRange]:
        """Detect the temporal range of the dataset.

        If no dataset is provided, it will be fetched according to the settings in the
        adapter. also see :py:meth:`hydromt.GeoDatasetAdapter.get_data`.

        Parameters
        ----------
        ds: xr.Dataset, xr.DataArray, Optional
            the dataset to detect the time range of. It must have a time dimentsion set.
            If none is provided, :py:meth:`hydromt.GeoDatasetAdapter.get_data`
            will be used to fetch the it before detecting.

        Returns
        -------
        range: Tuple[np.datetime64, np.datetime64]
            A tuple containing the start and end of the time dimension. Range is
            inclusive on both sides.
        """
        if ds is None:
            ds = self.get_data()
        return (
            ds[ds.vector.time_dim].min().values,
            ds[ds.vector.time_dim].max().values,
        )

    def to_stac_catalog(
        self,
        on_error: ErrorHandleMethod = ErrorHandleMethod.COERCE,
    ) -> Optional[StacCatalog]:
        """
        Convert a geodataset into a STAC Catalog representation.

        The collection will contain an asset for each of the associated files.


        Parameters
        ----------
        - on_error (str, optional): The error handling strategy.
          Options are: "raise" to raise an error on failure, "skip" to skip
          the dataset on failure, and "coerce" (default) to set default
          values on failure.

        Returns
        -------
        - Optional[StacCatalog]: The STAC Catalog representation of the dataset, or
          None if the dataset was skipped.
        """
        try:
            bbox, crs = self.get_bbox(detect=True)
            bbox = list(bbox)
            start_dt, end_dt = self.get_time_range(detect=True)
            start_dt = pd.to_datetime(start_dt)
            end_dt = pd.to_datetime(end_dt)
            props = {**self.meta, "crs": crs}
            ext = splitext(self.path)[-1]
            if ext in [".nc", ".vrt"]:
                media_type = MediaType.HDF5
            else:
                raise RuntimeError(
                    f"Unknown extention: {ext} cannot determine media type"
                )
        except (IndexError, KeyError, CRSError) as e:
            if on_error == ErrorHandleMethod.SKIP:
                logger.warning(
                    "Skipping {name} during stac conversion because"
                    "because detecting spacial extent failed."
                )
                return
            elif on_error == ErrorHandleMethod.COERCE:
                bbox = [0.0, 0.0, 0.0, 0.0]
                props = self.meta
                start_dt = datetime(1, 1, 1)
                end_dt = datetime(1, 1, 1)
                media_type = MediaType.JSON
            else:
                raise e

        stac_catalog = StacCatalog(
            self.name,
            description=self.name,
        )
        stac_item = StacItem(
            self.name,
            geometry=None,
            bbox=bbox,
            properties=props,
            datetime=None,
            start_datetime=start_dt,
            end_datetime=end_dt,
        )
        stac_asset = StacAsset(str(self.path), media_type=media_type)
        base_name = basename(self.path)
        stac_item.add_asset(base_name, stac_asset)

        stac_catalog.add_item(stac_item)
        return stac_catalog



# -*- coding: utf-8 -*-
"""HydroMT models API."""
from .. import _compat
from .model_api import *
from .model_grid import *
from .model_network import *
from .model_plugins import *
from .model_vector import *

if _compat.HAS_XUGRID:
    from .model_mesh import MeshModel

# expose global MODELS object which discovers and loads
# any local generalized or plugin model class on-the-fly
MODELS = ModelCatalog()



# -*- coding: utf-8 -*-
"""HydroMT GridModel class definition."""

import logging
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from affine import Affine
from pyproj import CRS
from shapely.geometry import box

from .. import gis_utils, raster, workflows
from .model_api import Model

__all__ = ["GridModel"]
logger = logging.getLogger(__name__)


class GridMixin(object):
    # placeholders
    # xr.Dataset representation of all static parameter maps at the same resolution and
    # bounds - renamed from staticmaps
    _API = {"grid": xr.Dataset}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._grid = None  # xr.Dataset()

    # generic grid methods
    def setup_grid_from_constant(
        self,
        constant: Union[int, float],
        name: str,
        dtype: Optional[str] = "float32",
        nodata: Optional[Union[int, float]] = None,
        mask_name: Optional[str] = "mask",
    ) -> List[str]:
        """HYDROMT CORE METHOD: Adds a grid based on a constant value.

        Parameters
        ----------
        constant: int, float
            Constant value to fill grid with.
        name: str
            Name of grid.
        dtype: str, optional
            Data type of grid. By default 'float32'.
        nodata: int, float, optional
            Nodata value. By default infered from dtype.
        mask_name: str, optional
            Name of mask in self.grid to use for masking raster_fn. By default 'mask'.
            Use None to disable masking.

        Returns
        -------
        list
            Names of added model grid layer.
        """
        da = workflows.grid.grid_from_constant(
            grid_like=self.grid,
            constant=constant,
            name=name,
            dtype=dtype,
            nodata=nodata,
            mask_name=mask_name,
        )
        # Add to grid
        self.set_grid(da)

        return [name]

    def setup_grid_from_rasterdataset(
        self,
        raster_fn: Union[str, Path, xr.DataArray, xr.Dataset],
        variables: Optional[List] = None,
        fill_method: Optional[str] = None,
        reproject_method: Optional[Union[List, str]] = "nearest",
        mask_name: Optional[str] = "mask",
        rename: Optional[Dict] = None,
    ) -> List[str]:
        """HYDROMT CORE METHOD: Add data variable(s) from ``raster_fn`` to grid object.

        If raster is a dataset, all variables will be added unless ``variables`` list
        is specified.

        Adds model layers:

        * **raster.name** grid: data from raster_fn

        Parameters
        ----------
        raster_fn: str, Path, xr.DataArray, xr.Dataset
            Data catalog key, path to raster file or raster xarray data object.
            If a path to a raster file is provided it will be added
            to the data_catalog with its name based on the file basename without
            extension.
        variables: list, optional
            List of variables to add to grid from raster_fn. By default all.
        fill_method : str, optional
            If specified, fills nodata values using fill_nodata method.
            Available methods are {'linear', 'nearest', 'cubic', 'rio_idw'}.
        reproject_method: list, str, optional
            See rasterio.warp.reproject for existing methods, by default 'nearest'.
            Can provide a list corresponding to ``variables``.
        mask_name: str, optional
            Name of mask in self.grid to use for masking raster_fn. By default 'mask'.
            Use None to disable masking.
        rename: dict, optional
            Dictionary to rename variable names in raster_fn before adding to grid
            {'name_in_raster_fn': 'name_in_grid'}. By default empty.

        Returns
        -------
        list
            Names of added model map layers
        """
        rename = rename or {}
        self.logger.info(f"Preparing grid data from raster source {raster_fn}")
        # Read raster data and select variables
        ds = self.data_catalog.get_rasterdataset(
            raster_fn,
            geom=self.region,
            buffer=2,
            variables=variables,
            single_var_as_array=False,
        )
        # Data resampling
        ds_out = workflows.grid.grid_from_rasterdataset(
            grid_like=self.grid,
            ds=ds,
            variables=variables,
            fill_method=fill_method,
            reproject_method=reproject_method,
            mask_name=mask_name,
            rename=rename,
        )
        # Add to grid
        self.set_grid(ds_out)

        return list(ds_out.data_vars.keys())

    def setup_grid_from_raster_reclass(
        self,
        raster_fn: Union[str, Path, xr.DataArray],
        reclass_table_fn: Union[str, Path, pd.DataFrame],
        reclass_variables: List,
        variable: Optional[str] = None,
        fill_method: Optional[str] = None,
        reproject_method: Optional[Union[List, str]] = "nearest",
        mask_name: Optional[str] = "mask",
        rename: Optional[Dict] = None,
        **kwargs,
    ) -> List[str]:
        """HYDROMT CORE METHOD: Add data variable(s) to grid object by reclassifying the data in ``raster_fn`` based on ``reclass_table_fn``.

        Adds model layers:

        * **reclass_variables** grid: reclassified raster data

        Parameters
        ----------
        raster_fn: str, Path, xr.DataArray
            Data catalog key, path to raster file or raster xarray data object.
            Should be a DataArray. Else use `variable` argument for selection.
        reclass_table_fn: str, Path, pd.DataFrame
            Data catalog key, path to tabular data file or tabular pandas dataframe
            object for the reclassification table of `raster_fn`.
        reclass_variables: list
            List of reclass_variables from reclass_table_fn table to add to maps.
            Index column should match values in `raster_fn`.
        variable: str, optional
            Name of raster_fn dataset variable to use. This is only required when
            reading datasets with multiple variables.
            By default None.
        fill_method : str, optional
            If specified, fills nodata values in `raster_fn` using fill_nodata method
            before reclassifying. Available methods are
            {'linear', 'nearest', 'cubic', 'rio_idw'}.
        reproject_method: str, optional
            See rasterio.warp.reproject for existing methods, by default "nearest".
            Can provide a list corresponding to ``reclass_variables``.
        mask_name: str, optional
            Name of mask in self.grid to use for masking raster_fn. By default 'mask'.
            Use None to disable masking.
        rename: dict, optional
            Dictionary to rename variable names in reclass_variables before adding to
            grid {'name_in_reclass_table': 'name_in_grid'}. By default empty.

        **kwargs : dict
            Additional keyword arguments to be passed to `get_rasterdataset`

        Returns
        -------
        list
            Names of added model grid layers
        """  # noqa: E501
        rename = rename or dict()
        self.logger.info(
            f"Preparing grid data by reclassifying the data in {raster_fn} based "
            f"on {reclass_table_fn}"
        )
        # Read raster data and remapping table
        da = self.data_catalog.get_rasterdataset(
            raster_fn, geom=self.region, buffer=2, variables=variable, **kwargs
        )
        if not isinstance(da, xr.DataArray):
            raise ValueError(
                f"raster_fn {raster_fn} should be a single variable. "
                "Please select one using the 'variable' argument"
            )
        df_vars = self.data_catalog.get_dataframe(
            reclass_table_fn, variables=reclass_variables
        )
        # Data resampling
        ds_vars = workflows.grid.grid_from_raster_reclass(
            grid_like=self.grid,
            da=da,
            reclass_table=df_vars,
            reclass_variables=reclass_variables,
            fill_method=fill_method,
            reproject_method=reproject_method,
            mask_name=mask_name,
            rename=rename,
        )
        # Add to maps
        self.set_grid(ds_vars.rename(rename))

        return list(ds_vars.data_vars.keys())

    def setup_grid_from_geodataframe(
        self,
        vector_fn: Union[str, Path, gpd.GeoDataFrame],
        variables: Optional[Union[List, str]] = None,
        nodata: Optional[Union[List, int, float]] = -1,
        rasterize_method: Optional[str] = "value",
        mask_name: Optional[str] = "mask",
        rename: Optional[Dict] = None,
        all_touched: Optional[bool] = True,
    ) -> List[str]:
        """HYDROMT CORE METHOD: Add data variable(s) to grid object by rasterizing the data from ``vector_fn``.

        Several type of rasterization are possible:
            * "fraction": the fraction of the grid cell covered by the vector
                shape is returned.
            * "area": the area of the grid cell covered by the vector shape is returned.
            * "value": the value from the variables columns of vector_fn are used.
                If this is used, variables must be specified.

        Parameters
        ----------
        vector_fn : str, Path, gpd.GeoDataFrame
            Data catalog key, path to vector file or a vector geopandas object.
        variables : List, str, optional
            List of variables to add to grid from vector_fn. Required if
            rasterize_method is "value", by default None.
        nodata : List, int, float, optional
            No data value to use for rasterization, by default -1. If a list is
            provided, it should have the same length has variables.
        rasterize_method : str, optional
            Method to rasterize the vector data. Either {"value", "fraction", "area"}.
            If "value", the value from the variables columns in vector_fn are used
            directly in the raster. If "fraction", the fraction of the grid
            cell covered by the vector file is returned. If "area", the area of the
            grid cell covered by the vector file is returned.
        mask_name: str, optional
            Name of mask in self.grid to use for masking raster_fn. By default 'mask'.
            Use None to disable masking.
        rename: dict, optional
            Dictionary to rename variable names in variables before adding to grid
            {'name_in_variables': 'name_in_grid'}. To rename with method fraction or
            area use {'vector_fn': 'name_in_grid'}. By default empty.
        all_touched : bool, optional
            If True (default), all pixels touched by geometries will be burned in.
            If false, only pixels whose center is within the polygon or that are
            selected by Bresenham's line algorithm will be burned in.

        Returns
        -------
        list
            Names of added model grid layers
        """  # noqa: E501
        rename = rename or dict()
        self.logger.info(f"Preparing grid data from vector '{vector_fn}'.")
        gdf = self.data_catalog.get_geodataframe(
            vector_fn, geom=self.region, dst_crs=self.crs
        )
        if gdf.empty:
            self.logger.warning(
                f"No shapes of {vector_fn} found within region,"
                " skipping setup_grid_from_vector."
            )
            return
        # Data resampling
        if vector_fn in rename.keys():
            # In case of choosing a new name with area or fraction method pass
            # the name directly
            rename = rename[vector_fn]
        ds = workflows.grid.grid_from_geodataframe(
            grid_like=self.grid,
            gdf=gdf,
            variables=variables,
            nodata=nodata,
            rasterize_method=rasterize_method,
            mask_name=mask_name,
            rename=rename,
            all_touched=all_touched,
        )
        # Add to grid
        self.set_grid(ds)

        return list(ds.data_vars.keys())

    @property
    def grid(self):
        """Model static gridded data as xarray.Dataset."""
        if self._grid is None:
            self._initialize_grid()
        return self._grid

    def _initialize_grid(self, skip_read=False) -> None:
        """Initialize grid object."""
        if self._grid is None:
            self._grid = xr.Dataset()
            if self._read and not skip_read:
                self.read_grid()

    def set_grid(
        self,
        data: Union[xr.DataArray, xr.Dataset, np.ndarray],
        name: Optional[str] = None,
    ):
        """Add data to grid.

        All layers of grid must have identical spatial coordinates.

        Parameters
        ----------
        data: xarray.DataArray or xarray.Dataset
            new map layer to add to grid
        name: str, optional
            Name of new map layer, this is used to overwrite the name of a DataArray
            and ignored if data is a Dataset
        """
        self._initialize_grid()
        # NOTE: variables in a dataset are not longer renamed as used to be the case in
        # set_staticmaps
        name_required = isinstance(data, np.ndarray) or (
            isinstance(data, xr.DataArray) and data.name is None
        )
        if name is None and name_required:
            raise ValueError(f"Unable to set {type(data).__name__} data without a name")
        if isinstance(data, np.ndarray):
            if data.shape != self.grid.raster.shape:
                raise ValueError("Shape of data and grid maps do not match")
            data = xr.DataArray(dims=self.grid.raster.dims, data=data, name=name)
        if isinstance(data, xr.DataArray):
            if name is not None:  # rename
                data.name = name
            data = data.to_dataset()
        elif not isinstance(data, xr.Dataset):
            raise ValueError(f"cannot set data of type {type(data).__name__}")

        if len(self._grid) == 0:  # empty grid
            self._grid = data
        else:
            for dvar in data.data_vars:
                if dvar in self._grid:
                    if self._read:
                        self.logger.warning(f"Replacing grid map: {dvar}")
                self._grid[dvar] = data[dvar]

    def read_grid(self, fn: str = "grid/grid.nc", **kwargs) -> None:
        """Read model grid data at <root>/<fn> and add to grid property.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.read_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, by default 'grid/grid.nc'
        **kwargs : dict
            Additional keyword arguments to be passed to the `read_nc` method.
        """
        self._assert_read_mode()
        self._initialize_grid(skip_read=True)

        # Load grid data in r+ mode to allow overwritting netcdf files
        if self._read and self._write:
            kwargs["load"] = True
        loaded_nc_files = self.read_nc(fn, single_var_as_array=False, **kwargs)
        for ds in loaded_nc_files.values():
            self.set_grid(ds)

    def write_grid(
        self,
        fn: str = "grid/grid.nc",
        gdal_compliant: bool = False,
        rename_dims: bool = False,
        force_sn: bool = False,
        **kwargs,
    ) -> None:
        """Write model grid data to netcdf file at <root>/<fn>.

        key-word arguments are passed to :py:meth:`~hydromt.models.Model.write_nc`

        Parameters
        ----------
        fn : str, optional
            filename relative to model root, by default 'grid/grid.nc'
        **kwargs : dict
            Additional keyword arguments to be passed to the `write_nc` method.
        gdal_compliant : bool, optional
            If True, write grid data in a way that is compatible with GDAL,
            by default False
        rename_dims: bool, optional
            If True and gdal_compliant, rename x_dim and y_dim to standard names
            depending on the CRS (x/y for projected and lat/lon for geographic).
        force_sn: bool, optional
            If True and gdal_compliant, forces the dataset to have
            South -> North orientation.
        """
        if len(self.grid) == 0:
            self.logger.debug("No grid data found, skip writing.")
        else:
            self._assert_write_mode()
            # write_nc requires dict - use dummy 'grid' key
            self.write_nc(
                {"grid": self.grid},
                fn,
                gdal_compliant=gdal_compliant,
                rename_dims=rename_dims,
                force_sn=force_sn,
                **kwargs,
            )


class GridModel(GridMixin, Model):

    """Model class Grid Model for gridded models in HydroMT."""

    _CLI_ARGS = {"region": "setup_grid"}
    _NAME = "grid_model"

    def __init__(
        self,
        root: str = None,
        mode: str = "w",
        config_fn: str = None,
        data_libs: List[str] = None,
        logger=logger,
    ):
        """Initialize a GridModel for distributed models with a regular grid."""
        super().__init__(
            root=root,
            mode=mode,
            config_fn=config_fn,
            data_libs=data_libs,
            logger=logger,
        )

    ## GENERIC SETUP METHODS
    def setup_grid(
        self,
        region: dict,
        res: Optional[float] = None,
        crs: int = None,
        rotated: bool = False,
        hydrography_fn: Optional[str] = None,
        basin_index_fn: Optional[str] = None,
        add_mask: bool = True,
        align: bool = True,
        dec_origin: int = 0,
        dec_rotation: int = 3,
    ) -> xr.DataArray:
        """HYDROMT CORE METHOD: Create a 2D regular grid or reads an existing grid.

        A 2D regular grid will be created from a geometry (geom_fn) or bbox. If an
        existing grid is given, then no new grid will be generated.

        Adds/Updates model layers (if add_mask):
        * **mask** grid mask: add grid mask to grid object

        Parameters
        ----------
        region : dict
            Dictionary describing region of interest, e.g.:
            * {'bbox': [xmin, ymin, xmax, ymax]}
            * {'geom': 'path/to/polygon_geometry'}
            * {'grid': 'path/to/grid_file'}
            * {'basin': [x, y]}

            Region must be of kind [grid, bbox, geom, basin, subbasin, interbasin].
        res: float
            Resolution used to generate 2D grid [unit of the CRS], required if region
            is not based on 'grid'.
        crs : EPSG code, int, str optional
            EPSG code of the model or "utm" to let hydromt find the closest projected
        rotated : bool, optional
            if True, a minimum rotated rectangular grid is fitted around the region,
            by default False. Only  applies if region is of kind 'bbox', 'geom'
        hydrography_fn : str
            Name of data source for hydrography data. Required if region is of kind
                'basin', 'subbasin' or 'interbasin'.

            * Required variables: ['flwdir'] and any other 'snapping' variable required
                to define the region.

            * Optional variables: ['basins'] if the `region` is based on a
                (sub)(inter)basins without a 'bounds' argument.

        basin_index_fn : str
            Name of data source with basin (bounding box) geometries associated with
            the 'basins' layer of `hydrography_fn`. Only required if the `region` is
            based on a (sub)(inter)basins without a 'bounds' argument.
        add_mask : bool, optional
            Add mask variable to grid object, by default True.
        align : bool, optional
            If True (default), align target transform to resolution.
        dec_origin : int, optional
            number of decimals to round the origin coordinates, by default 0
        dec_rotation : int, optional
            number of decimals to round the rotation angle, by default 3

        Returns
        -------
        grid : xr.DataArray
            Generated grid mask.
        """
        self.logger.info("Preparing 2D grid.")

        kind = next(iter(region))  # first key of region
        if kind in ["bbox", "geom", "basin", "subbasin", "interbasin"]:
            # Do not parse_region for grid as we want to allow for more (file) formats
            kind, region = workflows.parse_region(
                region, data_catalog=self.data_catalog, logger=self.logger
            )
        elif kind != "grid":
            raise ValueError(
                f"Region for grid must be of kind [grid, bbox, geom, basin, subbasin,"
                f" interbasin], kind {kind} not understood."
            )

        # Derive xcoords, ycoords and geom for the different kind options
        if kind in ["bbox", "geom"]:
            if not isinstance(res, (int, float)):
                raise ValueError("res argument required for kind 'bbox', 'geom'")
            if kind == "bbox":
                bbox = region["bbox"]
                geom = gpd.GeoDataFrame(geometry=[box(*bbox)], crs=4326)
            elif kind == "geom":
                geom = region["geom"]
                if geom.crs is None:
                    raise ValueError('Model region "geom" has no CRS')
            if crs is not None:
                crs = gis_utils.parse_crs(crs, bbox=geom.total_bounds)
                geom = geom.to_crs(crs)
            # Generate grid based on res for region bbox
            # TODO add warning on res value if crs is projected or not?
            if not rotated:
                xmin, ymin, xmax, ymax = geom.total_bounds
                res = abs(res)
                if align:
                    xmin = round(xmin / res) * res
                    ymin = round(ymin / res) * res
                    xmax = round(xmax / res) * res
                    ymax = round(ymax / res) * res
                xcoords = np.linspace(
                    xmin + res / 2,
                    xmax - res / 2,
                    num=round((xmax - xmin) / res),
                    endpoint=True,
                )
                ycoords = np.flip(
                    np.linspace(
                        ymin + res / 2,
                        ymax - res / 2,
                        num=round((ymax - ymin) / res),
                        endpoint=True,
                    )
                )
            else:  # rotated
                geomu = geom.unary_union
                x0, y0, mmax, nmax, rot = workflows.grid.rotated_grid(
                    geomu, res, dec_origin=dec_origin, dec_rotation=dec_rotation
                )
                transform = (
                    Affine.translation(x0, y0)
                    * Affine.rotation(rot)
                    * Affine.scale(res, res)
                )

        elif kind in ["basin", "subbasin", "interbasin"]:
            # retrieve global hydrography data (lazy!)
            ds_hyd = self.data_catalog.get_rasterdataset(hydrography_fn)
            if "bounds" not in region:
                region.update(basin_index=self.data_catalog.get_source(basin_index_fn))
            # get basin geometry
            geom, xy = workflows.get_basin_geometry(
                ds=ds_hyd,
                kind=kind,
                logger=self.logger,
                **region,
            )
            # get ds_hyd again but clipped to geom, one variable is enough
            da_hyd = self.data_catalog.get_rasterdataset(
                hydrography_fn, geom=geom, variables=["flwdir"]
            )
            if not isinstance(res, (int, float)):
                self.logger.info(
                    "res argument not defined, using resolution of "
                    f"hydrography_fn {da_hyd.raster.res}"
                )
                res = da_hyd.raster.res
            # Reproject da_hyd based on crs and grid and align, method is not important
            # only coords will be used
            # TODO add warning on res value if crs is projected or not?
            if res != da_hyd.raster.res and crs != da_hyd.raster.crs:
                da_hyd = da_hyd.raster.reproject(
                    dst_crs=crs,
                    dst_res=res,
                    align=align,
                )
            # Get xycoords, geom
            xcoords = da_hyd.raster.xcoords.values
            ycoords = da_hyd.raster.ycoords.values
            if geom.crs != da_hyd.raster.crs:
                crs = da_hyd.raster.crs
                geom = geom.to_crs(crs)
        elif kind == "grid":
            # Support more formats for grid input (netcdf, zarr, io.open_raster)
            fn = region[kind]
            if isinstance(fn, (xr.DataArray, xr.Dataset)):
                da_like = fn
            else:
                da_like = self.data_catalog.get_rasterdataset(fn)
            # Get xycoords, geom
            xcoords = da_like.raster.xcoords.values
            ycoords = da_like.raster.ycoords.values
            geom = da_like.raster.box
            if crs is not None or res is not None:
                self.logger.warning(
                    "For region kind 'grid', the gris crs/res are used and not"
                    f" user-defined crs {crs} or res {res}"
                )
            crs = da_like.raster.crs

        # Instantiate grid object
        # Generate grid using hydromt full method
        if not rotated:
            coords = {"y": ycoords, "x": xcoords}
            grid = raster.full(
                coords=coords,
                nodata=1,
                dtype=np.uint8,
                name="mask",
                attrs={},
                crs=geom.crs,
                lazy=False,
            )
        else:
            grid = raster.full_from_transform(
                transform,
                shape=(mmax, nmax),
                nodata=1,
                dtype=np.uint8,
                name="mask",
                attrs={},
                crs=geom.crs,
                lazy=False,
            )
        # Create geometry_mask with geom
        if add_mask:
            grid = grid.raster.geometry_mask(geom, all_touched=True)
            grid.name = "mask"
        # Remove mask variable mask from grid if not add_mask
        else:
            grid = grid.to_dataset()
            grid = grid.drop_vars("mask")

        # Add region and grid to model
        self.set_geoms(geom, "region")
        self.set_grid(grid)

        return grid

    ## I/O
    def read(
        self,
        components: List = None,
    ) -> None:
        """Read the complete model schematization and configuration from model files.

        Parameters
        ----------
        components : List, optional
            List of model components to read, each should have an associated
            read_<component> method. By default ['config', 'maps', 'grid',
            'geoms', 'tables', 'forcing', 'states', 'results']
        """
        components = components or [
            "config",
            "grid",
            "geoms",
            "tables",
            "forcing",
            "states",
            "results",
        ]
        super().read(components=components)

    def write(
        self,
        components: List = None,
    ) -> None:
        """Write the complete model schematization and configuration to model files.

        Parameters
        ----------
        components : List, optional
            List of model components to write, each should have an
            associated write_<component> method. By default
            ['config', 'maps', 'grid', 'geoms', 'tables', 'forcing', 'states']
        """
        components = components or [
            "config",
            "maps",
            "grid",
            "geoms",
            "tables",
            "forcing",
            "states",
        ]
        super().write(components=components)

    # Properties for subclass GridModel
    @property
    def res(self) -> Tuple[float, float]:
        """Returns the resolution of the model grid."""
        if len(self.grid) > 0:
            return self.grid.raster.res

    @property
    def transform(self):
        """Returns spatial transform of the model grid."""
        if len(self.grid) > 0:
            return self.grid.raster.transform

    @property
    def crs(self) -> Union[CRS, None]:
        """Returns coordinate reference system embedded in the model grid."""
        if self.grid.raster.crs is not None:
            return CRS(self.grid.raster.crs)

    @property
    def bounds(self) -> List[float]:
        """Returns the bounding box of the model grid."""
        if len(self.grid) > 0:
            return self.grid.raster.bounds

    @property
    def region(self) -> gpd.GeoDataFrame:
        """Returns the geometry of the model area of interest."""
        region = gpd.GeoDataFrame()
        if "region" in self.geoms:
            region = self.geoms["region"]
        elif len(self.grid) > 0:
            crs = self.crs
            if crs is not None and hasattr(crs, "to_epsg"):
                crs = crs.to_epsg()  # not all CRS have an EPSG code
            region = gpd.GeoDataFrame(geometry=[box(*self.bounds)], crs=crs)
        return region

    def set_crs(self, crs: CRS) -> None:
        """Set coordinate reference system of the model grid."""
        if len(self.grid) > 0:
            self.grid.raster.set_crs(crs)



"""Type aliases used by hydromt."""

from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Literal, Tuple, TypedDict, Union

from geopandas import GeoDataFrame, GeoSeries
from xarray import DataArray, Dataset

StrPath = Union[str, Path]
GeoDataframeSource = StrPath
GeoDatasetSource = StrPath
RasterDatasetSource = StrPath
DatasetSource = StrPath

Bbox = Tuple[float, float, float, float]
Crs = int
TotalBounds = Tuple[Bbox, Crs]
TimeRange = Tuple[datetime, datetime]
Number = Union[int, float]
SourceSpecDict = TypedDict(
    "SourceSpecDict", {"source": str, "provider": str, "version": Union[str, int]}
)

DeferedFileClose = TypedDict(
    "DeferedFileClose",
    {"ds": Dataset, "org_fn": str, "tmp_fn": str, "close_attempts": int},
)
XArrayDict = Dict[str, Union[DataArray, Dataset]]

ExportConfigDict = TypedDict(
    "ExportConfigDict",
    {"args": Dict[str, Any], "meta": Dict[str, Any], "sources": List[SourceSpecDict]},
)

Predicate = Literal[
    "intersects", "within", "contains", "overlaps", "crosses", "touches"
]

Geom = Union[GeoDataFrame, GeoSeries]

Data = Union[Dataset, DataArray]

Variables = Union[str, List[str]]

GeomBuffer = int


class ErrorHandleMethod(Enum):
    """Strategies for error handling within hydromt."""

    RAISE = 1
    SKIP = 2
    COERCE = 3



"""Pydantic models for the validation of region specifications."""
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field, model_validator


class PathRegion(BaseModel):
    """A validation model for a region loaded from a file."""

    path: Path

    @staticmethod
    def from_path(path: Union[Path, str]) -> "PathRegion":
        """Create a region that will be loaded from a file."""
        if isinstance(path, Path):
            return PathRegion(path=path)
        else:
            return PathRegion(path=Path(path))

    @model_validator(mode="after")
    def _check_path_exists(self) -> "PathRegion":
        if not self.path.exists():
            raise ValueError(f"Path not found at {self.path}")
        return self


class BoundingBoxRegion(BaseModel):
    """A validation model for a region described by a bounding box in WGS84 space."""

    xmin: float = Field(ge=-180)
    xmax: float = Field(le=180)
    ymin: float = Field(ge=-90)
    ymax: float = Field(le=90)

    @staticmethod
    def from_list(
        input: Union[Tuple[float, float, float, float], List[float]]
    ) -> "BoundingBoxRegion":
        """Create a region specification from a [xmin,ymin,xmax,ymax] list."""
        xmin, ymin, xmax, ymax = input
        return BoundingBoxRegion(
            xmin=xmin,
            ymin=ymin,
            xmax=xmax,
            ymax=ymax,
        )

    @staticmethod
    def from_dict(input_dict: Dict) -> "BoundingBoxRegion":
        """Create a region specification from dictionary specifying values for xmin, ymin, xmax, ymax."""
        xmin, ymin, xmax, ymax = input_dict["bbox"]

        return BoundingBoxRegion(
            xmin=xmin,
            ymin=ymin,
            xmax=xmax,
            ymax=ymax,
        )

    @model_validator(mode="after")
    def _check_bounds_ordering(self) -> "BoundingBoxRegion":
        # pydantic will turn these asserion errors into validation errors for us
        assert self.xmin <= self.xmax
        assert self.ymin <= self.ymax
        return self


Region = Union[
    BoundingBoxRegion,
    PathRegion,
]


def validate_region(input: Dict[str, Any]) -> Optional[Region]:
    """Create a validated model from a dictionary as allowed by the CLI."""
    if "bbox" in input:
        val = input["bbox"]
        if isinstance(val, list):
            return BoundingBoxRegion.from_list(val)
        elif isinstance(val, dict):
            return BoundingBoxRegion.from_dict(val)
    elif "geom" in input:
        val = input["geom"]
        return PathRegion.from_path(val)
    else:
        for region_type in ["grid", "mesh", "basin", "subbasin", "interbasin"]:
            if region_type in input:
                raise NotImplementedError(
                    f"region kind {region_type} is not supported in region validation yet, but is recognised by HydroMT."
                )

        raise NotImplementedError(f"Unknown region kind: {input}")



"""Implementation of data downloading functionality."""

import glob
import gzip
import os
from os.path import isdir

import numpy as np
import pandas as pd
import rasterio
import requests
import xarray as xr
from requests import HTTPError

FILE_DIR = os.path.dirname(__file__)


def download_file(url, outdir=FILE_DIR):
    """Download a file from the given URL and save it to the specified output directory.

    Args:
    ----
    url (str): The URL of the file to download.
    outdir (str): The output directory to save the downloaded file.

    Returns
    -------
        None
    """
    basename = url.split("/")[-1]
    local_filename = os.path.join(outdir, basename)
    if os.path.isfile(local_filename):
        return
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        print(f"downloading {basename} ..")
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                # If you have chunk encoded response uncomment if
                # and set chunk_size parameter to None.
                # if chunk:
                f.write(chunk)


def download_africa_daily(outroot, start="2020-01-01", end="2021-03-31"):
    """Download daily CHIRPS data for Africa within the specified date range.

    Arguments
    ---------
    outroot :
        The root directory to save the downloaded files.
    start :
        The start date in the format "YYYY-MM-DD" (default: "2020-01-01").
    end :
        The end date in the format "YYYY-MM-DD" (default: "2021-03-31").

    Returns
    -------
        None
    """
    BASE_URL = "https://data.chc.ucsb.edu/products/CHIRPS-2.0/africa_daily/tifs/p05"
    for date in pd.date_range(start=start, end=end, freq="d"):
        if date > pd.to_datetime("today").date():
            return
        date_str = date.strftime("%Y.%m.%d")
        year = date.year
        outdir = os.path.join(outroot, str(year))
        if not isdir(outdir):
            os.makedirs(outdir)
        basename = f"chirps-v2.0.{date_str}.tif.gz"
        url = f"{BASE_URL}/{year}/{basename}"
        try:
            download_file(url, outdir)
        except HTTPError:
            fn_out = os.path.join(outdir, basename)
            if os.path.isfile(fn_out):
                os.unlink(fn_out)
            try:  # without .gz
                download_file(url[:-3], outdir)
            except HTTPError as r:
                print(r)
                fn_out = os.path.join(outdir, basename[:-3])
                if os.path.isfile(fn_out):
                    os.unlink(fn_out)
                continue


def tifs_to_nc(folder, year):
    """Convert downloaded CHIRPS TIFF files for a specific year to NetCDF format.

    Arguments
    ---------
    folder: str
        The root folder where the downloaded TIFF files are located.
    year: int
        The year to convert to NetCDF.

    """
    path = os.path.join(folder, str(year), f"chirps-v2.0.{year}*.tif*")
    nodata = -9999.0
    files = sorted(glob.glob(path))
    fn_out = os.path.join(folder, f"CHIRPS_rainfall_{year}.nc")
    if len(files) == 0 or os.path.isfile(fn_out):
        return

    for i, f in enumerate(files):
        infile = gzip.open(f) if f.endswith("gz") else f
        with rasterio.open(infile) as src:
            print(f"processing file : {os.path.basename(f)}")
            data = src.read(1)
            src_profile = src.profile
            out_transform = src_profile["transform"]

            nx, ny = data.shape[1], data.shape[0]
            x = (np.arange(nx) + 0.5) * out_transform[0] + out_transform[2]
            y = (np.arange(ny) + 0.5) * out_transform[4] + out_transform[5]

            time = "-".join(os.path.basename(f).split(".")[2:5])

            if i == 0:
                t = pd.to_datetime([time])
                ds = np.reshape(data, (1, ny, nx))
            else:
                t = pd.DatetimeIndex.append(t, pd.to_datetime([time]))
                ds = np.append(ds, np.reshape(data, (1, ny, nx)), axis=0)
        if f.endswith("gz"):
            infile.close()

    ds_img = xr.Dataset(
        {"precipitation": (["time", "lat", "lon"], ds)},
        coords={"lon": x, "lat": y, "time": t},
    )
    ds_img.precipitation.attrs["standard_name"] = "precipitation"
    ds_img.precipitation.attrs["units"] = "mm"
    ds_img.lon.attrs["standard_name"] = "longitude"
    ds_img.lon.attrs["units"] = "degrees _east"
    ds_img.lon.attrs["axis"] = "X"
    ds_img.lat.attrs["standard_name"] = "latitude"
    ds_img.lat.attrs["units"] = "degrees _north"
    ds_img.lat.attrs["axis"] = "Y"
    encoding = {
        "precipitation": {
            "_FillValue": nodata,
            "complevel": 4,
            "zlib": True,
            "chunksizes": [1, 320, 300],
        },
        "time": {"chunksizes": [90]},
    }
    print(f"writing to {fn_out} ..")
    ds_img.to_netcdf(fn_out, engine="netcdf4", encoding=encoding, unlimited_dims="time")


if __name__ == "__main__":
    outroot = r"p:\wflow_global\hydromt_staging\chirps"

    for year in range(2022, 2023):
        download_africa_daily(outroot, start=f"{year}-01-01", end=f"{year}-12-31")
        tifs_to_nc(outroot, year)



"""Utility functions for hydromt that have no other home."""

from typing import Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr


class _classproperty(property):
    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


def partition_dictionaries(left, right):
    """Calculate a partitioning of the two dictionaries.

    given dictionaries A and B this function will the follwing partition:
    (A ∩ B, A - B, B - A)
    """
    common = {}
    left_less_right = {}
    right_less_left = {}
    key_union = set(left.keys()) | set(right.keys())

    for key in key_union:
        value_left = left.get(key, None)
        value_right = right.get(key, None)
        if isinstance(value_left, dict) and isinstance(value_right, dict):
            (
                common_children,
                unique_left_children,
                unique_right_children,
            ) = partition_dictionaries(value_left, value_right)
            common[key] = common_children
            if unique_left_children != unique_right_children:
                left_less_right[key] = unique_left_children
                right_less_left[key] = unique_right_children
        elif value_left == value_right:
            common[key] = value_left
        else:
            if value_left is not None:
                left_less_right[key] = value_left
            if value_right is not None:
                right_less_left[key] = value_right

    return common, left_less_right, right_less_left


def elevation2rgba(val, nodata=np.nan):
    """Convert elevation to rgb tuple."""
    val += 32768
    r = np.floor(val / 256).astype(np.uint8)
    g = np.floor(val % 256).astype(np.uint8)
    b = np.floor((val - np.floor(val)) * 256).astype(np.uint8)
    mask = np.isnan(val) if np.isnan(nodata) else val == nodata
    a = np.where(mask, 0, 255).astype(np.uint8)
    return np.stack((r, g, b, a), axis=2)


def rgba2elevation(rgba: np.ndarray, nodata=np.nan, dtype=np.float32):
    """Convert rgb tuple to elevation."""
    r, g, b, a = np.split(rgba, 4, axis=2)
    val = (r * 256 + g + b / 256) - 32768
    return np.where(a == 0, nodata, val).squeeze().astype(dtype)


def _dict_pprint(d):
    import json

    return json.dumps(d, indent=2)


def has_no_data(
    data: Optional[Union[pd.DataFrame, gpd.GeoDataFrame, xr.Dataset, xr.DataArray]],
) -> bool:
    """Check whether various data containers are empty."""
    if data is None:
        return True
    elif isinstance(data, xr.Dataset):
        return all([v.size == 0 for v in data.data_vars.values()])
    else:
        return len(data) == 0



"""A simple script to generate enviroment.yml files from pyproject.toml."""

import argparse
import re
from sys import version_info
from typing import List

if version_info.minor >= 11:
    from tomllib import load
else:
    from tomli import load


# our quick and dirty implementation of recursive depedencies
def _parse_profile(profile_str: str, opt_deps: dict, project_name: str) -> List[str]:
    if profile_str is None or profile_str == "":
        return []

    pat = re.compile(r"\s*" + project_name + r"\[(.*)\]\s*")
    parsed = []
    queue = [f"{project_name}[{x.strip()}]" for x in profile_str.split(",")]
    while len(queue) > 0:
        dep = queue.pop(0)
        if dep == "":
            continue
        m = pat.match(dep)
        if m:
            # if we match the patern, all list elts have to be dependenciy groups
            dep_groups = [d.strip() for d in m.groups(0)[0].split(",")]
            unknown_dep_groups = set(dep_groups) - set(opt_deps.keys())
            if len(unknown_dep_groups) > 0:
                raise RuntimeError(f"unknown dependency group(s): {unknown_dep_groups}")
            queue.extend(dep_groups)
            continue

        if dep in opt_deps:
            queue.extend([x.strip() for x in opt_deps[dep]])
        else:
            parsed.append(dep)

    return parsed


parser = argparse.ArgumentParser()

parser.add_argument("profile", default="dev,test", nargs="?")
parser.add_argument("--output", "-o", default="environment.yml")
parser.add_argument("--channels", "-c", default=None)
parser.add_argument("--name", "-n", default=None)
parser.add_argument(
    "--py-version", "-p", default="3.11", choices=["3.9", "3.10", "3.11"]
)
parser.add_argument("-r", "--release", action="store_true")
args = parser.parse_args()

#
with open("pyproject.toml", "rb") as f:
    toml = load(f)
deps = toml["project"]["dependencies"]
opt_deps = toml["project"]["optional-dependencies"]
project_name = toml["project"]["name"]
# specific conda_install settings
install_config = toml["tool"].get("make_env", {})
deps_not_in_conda = install_config.get("deps_not_in_conda", [])
channels = install_config.get("channels", ["conda-forge"])
if args.channels is not None:
    channels.extend(args.channels.split(","))
    channels = list(set(channels))

# parse environment name
name = args.name
if name is None:
    name = project_name
print(f"Environment name: {name}")

# parse dependencies groups and flavours
# "min" equals no optional dependencies
if args.release:
    deps_to_install = ["hydromt"]
else:
    deps_to_install = deps.copy()
if args.profile not in ["", "min"]:
    extra_deps = _parse_profile(args.profile, opt_deps, project_name)
    deps_to_install.extend(extra_deps)

conda_deps = []
pip_deps = []
for dep in deps_to_install:
    if dep in deps_not_in_conda:
        pip_deps.append(dep)
    else:
        conda_deps.append(dep)
if args.py_version is not None:
    conda_deps.append(f"python~={args.py_version}.0")

# add pip as a conda dependency if we have pip deps
if len(pip_deps) > 0:
    conda_deps.append("pip")

# the list(set()) is to remove duplicates
conda_deps_to_install_string = "\n- ".join(sorted(list(set(conda_deps))))
channels_string = "\n- ".join(set(channels))

# create environment.yml
env_spec = f"""name: {name}

channels:
- {channels_string}

dependencies:
- {conda_deps_to_install_string}
"""
if len(pip_deps) > 0:
    pip_deps_to_install_string = "\n  - ".join(sorted(list(set(pip_deps))))
    env_spec += f"""- pip:
  - {pip_deps_to_install_string}
"""

with open(args.output, "w") as out:
    out.write(env_spec)
